"""Collection of utilities of the freva command line argument parser."""

from __future__ import annotations

import abc
import argparse
import logging
import os
import sys
import sysconfig
from copy import copy
from pathlib import Path
from typing import Callable, Optional, Type

import lazy_import

from evaluation_system.misc import FrevaLogger, logger

freva = lazy_import.lazy_module("freva")
config = lazy_import.lazy_module("evaluation_system.misc.config")


def get_cli_class(name: str) -> Optional[Type[BaseParser]]:
    """Get core module or cli extension.

    Parameters
    ----------
    name: str
        name of the cli sub module that is to be imported. This can be
        either a cli sub module for the freva core (databrowser, plugin etc)
        or an extension. Extensions are assumed to follow the following naming
        conventions: :py:mod:`<name>.app`. The ``app`` sub module must contain
        a class named :py:class:`Cli`.

    Returns
    -------
    module: The imported module
    """
    mod_name = name.replace("-", "_")
    try:  # First try importing the freva core functionality
        mod = __import__(f"freva.cli.{mod_name}", fromlist=[""])
    except ImportError:
        try:  # Then additional modules
            mod = __import__(f"{mod_name}.cli", fromlist=[""])
        except ImportError:
            return None
    if hasattr(mod, "Cli") and hasattr(mod.Cli, "desc"):
        return mod.Cli
    return None


class BaseParser(metaclass=abc.ABCMeta):
    """Base class that is used to construct a valid freva cli tool.

    This class can be used to extent the freva commands with additional sub
    commands.

    Parameters
    ----------
    parser: argparse.ArgumentParser, default: None
        Already parsed arguments if, None given (default) a new
        :py:class:`ArgumentParser` will should be constructed.

    Attributes
    ----------
    desc: str
        The class should have the :py:attribute:`desc` attribute set. This
        attribute is used to display a short summary over what this command
        is supposed to do.


    Example
    -------
    Suppose your tool named ``mytool`` should be made a freva cli extension the
    then the ``mytool`` library should define a command line interface that
    resides in the ``mytool.cli`` sub module of ``mytool``. Additionally your
    ``mytool.cli`` sub module has to define a class called ``Cli``. The content
    of the ``Cli`` class could like the following:

    .. code-block:: python

        import argparse
        from pathlib import Path
        import sys
        from typing import Optional

        from freva.cli import BaseParser

        class Cli(BaseParser):

            desc = "Apply mytool."

            def __init__(parser: Optional[argparse.ArgumentParser] = None):

                super().__init__("mytool", parser)
                self.parser.add_argument("path",
                                         type=Path,
                                         help="The first argument"
                )

            def do_something(path: Path) -> list[str]:
                '''Do the actual work, for example list a directory.'''
                return sorted(path.iterdir())

        def main():
            '''The main method calling the cli from a stand alone tool.'''

            arg_p = Cli()
            args = arg_p.parse_args(sys.argv[1:])
            arg_p.do_something(args.path)

    Following this outline all that would be needed to do is creating an entry
    point named ``freva-mytool`` pointing to ``mytool.cli:main``. This would
    be sufficient to create a freva mytool sub command.
    """

    desc: str = ""
    """The short description of a freva command."""

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        command: str = "freva",
    ):
        self.logger.is_cli = True
        self.parser = parser or argparse.ArgumentParser(
            prog=command,
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

    @property
    def logger(self) -> FrevaLogger:
        """Use evaluation_system logger in all classes using this class."""
        return logger

    def set_debug(self, debug: bool) -> None:
        """Set the logger level to DEBUG."""
        if debug is True:
            self.logger.setLevel(logging.DEBUG)

    def parse_args(self, argv: Optional[list[str]] = None) -> argparse.Namespace:
        """Parse the command line arguments."""
        args, unknown = self.parser.parse_known_args(argv)
        for arg in unknown:
            if arg.startswith("-"):
                self.parser.error(f"Unknown option: {arg}")
        self.kwargs = {k: v for (k, v) in args._get_kwargs() if k != "apply_func"}
        if unknown:
            self.kwargs["unknown"] = unknown
        self.set_debug(self.kwargs.pop("debug", False))
        return args


class SubCommandParser(BaseParser):
    """Base class for all freva sub command line argument parsers."""

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        sub_parsers: Optional[dict[str, Type[BaseParser]]] = None,
        command: str = "freva",
    ) -> None:
        """Create the sub-command parsers."""
        super().__init__(parser, command)
        self.subparsers = self.parser.add_subparsers(help="Available sub-commands:")
        self.sub_commands = sub_parsers or self.get_subcommand_parsers()
        self.help = self.get_subcommand_help(self.sub_commands)
        for cmd, subparser in self.sub_commands.items():
            self.parse_subcommand(cmd, subparser)

    def parse_subcommand(self, command: str, cli_class: Type[BaseParser]) -> None:
        """Parse the databrowser command."""

        call_parsers = self.subparsers.add_parser(
            command,
            description=cli_class.desc,
            help=cli_class.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        cli_class(call_parsers)

    def _usage(self, *args: Optional[str], **kwargs: Optional[str]) -> None:
        """Exit with usage message."""
        self.parser.error(
            "the following sub-commands are "
            f"required: {', '.join(self.sub_commands.keys())}"
        )

    @staticmethod
    def get_subcommand_parsers() -> dict[str, Type[BaseParser]]:
        """Create the help strings of the available sub commands."""
        path_list = [
            Path(p)
            for p in (os.environ.get("PATH") or os.defpath).split(os.pathsep)
            + [sysconfig.get_paths().get("scripts", "")]
            if p and Path(p).is_dir()
        ]
        parsers = {}
        for _dir in path_list:
            for file in _dir.iterdir():
                if file.name.startswith("freva-"):
                    _, _, sub_cmd = file.name.partition("-")
                    if sub_cmd not in parsers:
                        parser_class = get_cli_class(sub_cmd)
                        if parser_class is not None:
                            parsers[sub_cmd] = parser_class
        return parsers

    @classmethod
    def get_subcommand_help(
        cls,
        parsers: Optional[dict[str, Type[BaseParser]]] = None,
    ) -> dict[str, str]:
        """Create the help strings of the available sub commands."""
        parsers = parsers or cls.get_subcommand_parsers()
        return {h: mod.desc for (h, mod) in parsers.items() if hasattr(mod, "desc")}


class BaseCompleter:
    """Base class for command line argument completers."""

    def __init__(
        self,
        metavar: str,
        argv: list[str],
        choices: Optional[dict[str, tuple[str, str]]] = None,
        shell: str = "bash",
        strip: bool = False,
        flags_only: bool = False,
    ):
        self.choices = choices or {}
        self.strip = strip
        self.metavar = metavar
        self.argv = argv
        self.flags_only = flags_only
        if shell == "zsh":
            self.get_print = self._print_zsh
        elif shell == "fish":
            self.get_print = self._print_fish
        else:
            self.get_print = self._print_default

    def _print_zsh(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (_help, func) in choices.items():
            if self.metavar != "databrowser" or key.startswith("-"):
                out.append(f"{key}[{_help}]{func}")
            else:
                out.append(f"{key}: {_help}")
        return out

    def _print_fish(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (_help, func) in choices.items():
            out.append(f"{key}: {_help}")
        return out

    def _print_default(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (_help, _) in choices.items():
            if self.metavar == "databrowser" and not key.startswith("-"):
                out.append(f"{key}: {_help}")
            else:
                out.append(key)
        return out

    def _get_databrowser_choices(self) -> dict[str, tuple[str, str]]:
        """Get the choices for databrowser command."""

        facet_args = []
        for arg in self.argv:
            if len(arg.split("=")) == 2:
                facet_args.append(arg)
        facets = BaseCompleter.arg_to_dict(facet_args)
        search = freva.facet_search(**facets)
        choices = {}
        for att, values in search.items():
            if att not in facets:
                keys = ",".join([v for n, v in enumerate(values)])
                choices[att] = (keys, "")
        return choices

    def _get_plugin_choices(self) -> dict[str, tuple[str, str]]:
        """Get the choices for the plugin command."""

        docs: dict[str, dict[str, str]] = {}
        desc: dict[str, dict[str, str]] = {}
        plugins: dict[str, str] = {}
        try:
            plugin_cache = freva.read_plugin_cache()
        except Exception:
            plugin_cache = []
        for plugin in plugin_cache:
            plugins[plugin.name] = plugin.description
            docs[plugin.name] = {}
            desc[plugin.name] = {}
            for param, doc, value in plugin.parameters:
                desc[plugin.name][param] = doc
                docs[plugin.name][param] = value
        args = [arg for arg in self.argv if not arg.startswith("-") and arg != "plugin"]
        choices = {}
        if not args:
            # No plugin name was given
            return {pl: (help, "") for pl, help in plugins.items()}
        try:
            plugin_config = docs[args[0]]
            setup = desc[args[0]]
        except KeyError:
            # Wrong plugin name
            return {}
        options = []
        for key in plugin_config.keys():
            option_present = False
            for arg in args:
                if arg.startswith(f"{key}="):
                    option_present = True
                    break
            if not option_present:
                options.append(f"{key}=")
        for option in options:
            opt = option.strip("=")
            if "file" in setup[opt].lower() or "dir" in setup[opt].lower():
                choices[opt] = (setup[opt], ":file:_files")
            else:
                choices[opt] = (setup[opt], "")
        return choices

    @property
    def command_choices(self) -> dict[str, tuple[str, str]]:
        """Get the command line arguments for all sub commands."""

        choices = {}
        if self.flags_only:
            return self.choices
        if self.metavar == "databrowser":
            choices = self._get_databrowser_choices()
        elif self.metavar == "plugin":
            choices = self._get_plugin_choices()
        return {**self.choices, **choices}

    def formatted_print(self) -> None:
        """Print all choices to be processed by the shell completion function."""

        out = self.get_print(self.command_choices)
        for line in out:
            if line.startswith("-") and self.strip and not self.flags_only:
                continue
            elif not line.startswith("-") and self.flags_only:
                continue
            print(line)

    @staticmethod
    def arg_to_dict(
        args: Optional[list[str]], append: bool = False
    ) -> dict[str, list[str]]:
        """Convert a parsed arguments with ``key=value`` pairs to a dictionary.

        Parameters:
        ----------
        args:
            Collection of arguments that are converted to dict, the arguments
            should be of form ``key=value``.
        append:
            If a key occurs twice convert first value to list and append second
            value to list, the default behaviour is updating the entry with
            the second value.
        Returns:
        --------
        dict: Dictionary representation of ``key=value`` pairs.
        """
        out_dict: dict[str, list[str]] = {}
        for arg in args or []:
            key, _, value = arg.partition("=")
            if append and key in out_dict:
                out_dict[key].append(value)
            else:
                out_dict[key] = [value]
        return out_dict

    @classmethod
    def _get_choices_from_parser(
        cls,
        parser: argparse.ArgumentParser,
        argv: list[str],
    ) -> dict[str, tuple[str, str]]:
        choices: dict[str, tuple[str, str]] = {}
        for action in parser._actions:
            action_type = ""
            if isinstance(action, argparse._SubParsersAction):
                if argv:
                    cmd = argv.pop(0)
                    sub_parser = action.choices.get(cmd, "")
                    if sub_parser:
                        choices.update(cls._get_choices_from_parser(sub_parser, argv))
                else:
                    for ch, _parser in action.choices.items():
                        choices[ch] = _parser.description or "", action_type
            if action.help == argparse.SUPPRESS:
                # This is an option that is not exposed to users
                continue
            if action.type == Path:
                action_type = ":file:_files"
            if action.option_strings:
                choice = action.option_strings[0]
            else:
                choice = action.dest
            if choice.lower() not in (
                "facet",
                "facets",
                "tool-name",
                "options",
                "tool-options",
                "==suppress==",
                "-h",
            ):
                choices[choice] = (action.help or "", action_type)
        return choices

    @classmethod
    def get_args_of_subcommand(cls, argv: list[str]) -> dict[str, tuple[str, str]]:
        """Get all possible arguments from a freva sub-command."""
        sub_command = argv.pop(0)
        parser_class = get_cli_class(sub_command)
        if parser_class is None:
            return {}
        cli_parser = parser_class()
        return cls._get_choices_from_parser(cli_parser.parser, argv)

    @classmethod
    def parse_choices(cls, argv: list[str]) -> BaseCompleter:
        """Create the completion choices from given cmd arguments."""
        parser = argparse.ArgumentParser(
            description="Get choices for command line arguments"
        )
        parser.add_argument(
            "command", help="First command, freva, freva-history etc.", type=str
        )
        parser.add_argument("--shell", help="Shell in use", default="bash", type=str)
        parser.add_argument(
            "--strip",
            help="Do not print options starting with - .",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--flags-only",
            help="Only print options starting with - .",
            default=False,
            action="store_true",
        )
        app, args = parser.parse_known_args(argv)
        main_choices = {
            k: (v, "") for k, v in SubCommandParser.get_subcommand_help().items()
        }
        if app.command == "freva" and not args:
            return cls(app.command, [], main_choices, shell=app.shell, strip=app.strip)
        if app.command != "freva":
            sub_cmd = "-".join(app.command.split("-")[1:])
            args.insert(0, sub_cmd)
        choices = {}
        if args[0] in main_choices:
            choices = cls.get_args_of_subcommand(copy(args))
        choices = {k: v for (k, v) in choices.items() if k not in args}
        return cls(
            args[0],
            args,
            choices,
            shell=app.shell,
            strip=app.strip,
            flags_only=app.flags_only,
        )


def print_choices(arguments: list[str]) -> None:
    """Print completion choices based of command line arguments.

    Parameters:
    -----------
    arguments:
        List of command line arguments, that are evaluated by the choice printer.
    """
    argv = [arg.strip() for arg in arguments if arg.strip()]
    comp = BaseCompleter.parse_choices(argv)
    comp.formatted_print()


def standard_main(
    cli_class: Callable, version: str, argv: Optional[list[str]] = None
) -> None:
    """Wrapper for entry point script.

    Parameters:
    -----------
    cli_class:
        The class that is used to parse the command line arguments.
    version:
        The version of the freva command line tool.
    argv:
        The command line arguments that are passed to the command line parser.
    """
    cli = cli_class()
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=version),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)
    except Exception as error:
        freva.utils.exception_handler(error, True)



"""Additional utilities."""

import json
import logging
import os
import shlex
import subprocess
import time
from fnmatch import fnmatch
from functools import wraps
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

try:
    from IPython import get_ipython
except ImportError:  # pragma: no cover
    get_ipython = lambda: None  # pragma: no cover

import lazy_import
from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.db.utils import OperationalError
from rich.console import Console
from rich.live import Live
from rich.spinner import Spinner

import freva
from evaluation_system.misc import logger
from evaluation_system.misc.exceptions import ConfigurationException
from evaluation_system.misc.utils import metadict as meta_type

pm = lazy_import.lazy_module("evaluation_system.api.plugin_manager")
cancel_command = lazy_import.lazy_callable(
    "evaluation_system.api.workload_manager.cancel_command"
)
cfg = lazy_import.lazy_module("evaluation_system.misc.config")
db_settings = lazy_import.lazy_module("evaluation_system.settings.database")


def is_jupyter() -> bool:
    """Determine if we're running within an IPython kernel

    taken from: https://stackoverflow.com/questions/34091701/determine-if-were-in-an-ipython-notebook-session


    >>> is_jupyter()
    False
    """
    # check for `kernel` attribute on the IPython instance
    return getattr(get_ipython(), "kernel", None) is not None


def handled_exception(func: Callable[..., Any]) -> Callable[..., Any]:
    """Wrap the exception handler around a function."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Wrapper function that handles the exception."""
        wrong_config_msg = {
            False: (
                "consult the `freva.config` class to see how to pass a "
                "valid configuration."
            ),
            True: (
                "export the EVALUATION_SYSTEM_CONFIG_FILE "
                "environment variable to set a valid configuration file."
            ),
        }
        try:
            return func(*args, **kwargs)
        except (ImproperlyConfigured, OperationalError):  # prgama: no cover
            # Wrap django error in a more informative exception
            msg = "Your freva instance doesn't seem to be properly configured: "
            raise ConfigurationException(
                msg + wrong_config_msg[logger.is_cli]
            ) from None
        except BaseException as error:
            exception_handler(error)

    return wrapper


def exception_handler(exception: BaseException, cli: bool = False) -> None:
    """Handle raising exceptions appropriately."""

    trace_back = exception.__traceback__
    appendix = {
        False: " - decrease log level via `freva.logger.setLevel(10)`",
        True: " - increase verbosity flags (-v)",
    }[cli]
    append_msg = ""
    # Set only the last traceback of the exception
    if logger.level > logging.DEBUG and trace_back is not None:
        last_trace_back = trace_back
        while last_trace_back.tb_next:
            last_trace_back = last_trace_back.tb_next
        exception.__traceback__ = TracebackType(
            tb_next=None,
            tb_frame=last_trace_back.tb_frame,
            tb_lineno=last_trace_back.tb_lineno,
            tb_lasti=last_trace_back.tb_lasti,
        )
        append_msg = f"{appendix} for more information"
    msg = str(exception) + append_msg
    if cli:
        logger.exception(msg, exc_info=exception)
        raise SystemExit
    if logger.is_cli is False:
        logger.error(msg)
    if logger.level > logging.DEBUG:
        raise exception from None
    raise exception


class PluginStatus:
    """A class to interact with the status of a plugin application.

    With help of this class you can:

        - Check if a plugin is still running.
        - Get all results (data or plot files) of a plugin.
        - Check the configuration of a plugin.
        - Wait until the plugin is finished.

    Example
    -------

    The output of the ``freva.run_plugin`` method is an instance of the
    ``PluginStatus`` class. That means you can directly use the output of
    :py:meth:``freva.run_plugin`` to interact with the plugin status:

    .. execute_code::

        import freva
        res = freva.run_plugin("dummypluginfolders")
        print(res.status)

    You can also create an instance of the class yourself, if you know the
    ``history_id`` of a specific plugin run. Note that you can query these ids
    by making use of the :py:meth:``freva.history`` method:

    .. execute_code::

        import freva
        # Get the last run of the dummypluginfolders plugin
        hist = freva.history(plugin="dummypluginfolders", limit=1)[:-1]
        res = freva.PluginStatus(hist["id"])
        print(res.status)

    Note: You won't be able to kill running jobs associated to history id's from other users.
    """

    def __init__(self, history_id: int) -> None:
        self._id: int = history_id

    def __repr__(self) -> str:
        return (
            f"PluginStatus('{self.plugin}', "
            f"config={str(self.configuration)}, status={self.status})"
        )

    @property
    def _hist(self) -> Dict[str, Any]:
        log_level = logger.level
        logger.setLevel(logging.WARNING)
        try:
            hist = cast(
                List[Dict[str, Any]],
                freva.history(entry_ids=self._id, return_results=True, user_name="*"),
            )[0]
        except IndexError:
            logger.setLevel(log_level)
            return {}
        finally:
            logger.setLevel(log_level)
        hist["batch_settings"] = pm.get_batch_settings(self._id)
        return hist

    @property
    def history_id(self) -> int:
        """Get the ID of this plugin run in the freva history."""
        return self._id

    @property
    def status(self) -> str:
        """Get the state of the current plugin run."""
        hist = self._hist
        status_dict = hist.get("status_dict", {})
        status = hist.get("status")
        return cast(str, status_dict.get(status, "unknown"))

    @property
    def configuration(self) -> Dict[str, Any]:
        """Get the plugin configuration."""
        return self._hist.get("configuration", {})

    @property
    def stdout(self) -> str:
        """Get the stdout of the plugin.

        Example
        -------
        Read the output of the plugin:

        .. execute_code::

            import freva
            res = freva.run_plugin("dummypluginfolders")
            print(res.stdout)

        """
        try:
            return Path(self._hist["slurm_output"]).read_text(encoding="utf-8")
        except (FileNotFoundError, KeyError):
            return ""

    @property
    def batch_id(self) -> Optional[int]:
        """Get the id of the batch job, if the plugin was a batchmode job."""
        return self._hist.get("batch_settings", {}).get("job_id")

    @property
    def job_script(self) -> str:
        """Get the content of the job_script, if it was a batchmode job."""
        return self._hist.get("batch_settings", {}).get("job_script", "")

    def kill(self) -> None:
        """Kill a running batch job.

        This method has only affect on jobs there have been submitted using the
        ``batchmode=True`` flag.
        """
        if self.batch_id is not None:
            kill_cmd = cancel_command(
                self._hist["batch_settings"]["workload_manager"], self.batch_id
            )
            try:
                _ = subprocess.run(
                    shlex.split(kill_cmd),
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    check=True,
                )
            except (subprocess.CalledProcessError, FileNotFoundError):
                logger.warning("Could not kill job with id %i", self.batch_id)

    @property
    def plugin(self) -> str:
        """Get the plugin name."""
        return self._hist.get("tool", "")

    def __str__(self) -> str:
        return json.dumps(self._hist, indent=3)

    @property
    def version(self) -> Tuple[int, int, int]:
        """Get the version of the plugin."""
        try:
            return cast(
                Tuple[int, int, int],
                tuple(
                    int(v.strip().strip("(").strip(")"))
                    for v in self._hist["version"].split(",")
                ),
            )
        except (KeyError, ValueError, TypeError):
            return (0, 0, 0)

    def wait(self, timeout: Union[float, int] = 28800) -> None:
        """Wait for a plugin to finish.

        This method will block until the plugin is running.

        Parameters
        ----------
        timeout: int, default: 28800
            Wait ``timeout`` seconds for the plugin to finish. If the plugin
            hasn't been finish raise a ValueError.

        Raises
        ------
        ValueError: If the plugin took longer than ``timeout`` seconds to finish.

        Example
        -------

        This can be useful if a plugin was started using the ``batchmode=True``
        option and the execution of the code should wait until the plugin is
        finished.


        .. execute_code::

            import freva
            res = freva.run_plugin("dummypluginfolders", batchmode=True)
            res.wait(timeout=60) # Give the plugin 60 seconds to finish.

        """
        dt, n_itt = 0.5, 0
        max_itt = float(timeout) / dt
        text = "Waiting for plugin to finish... "
        spinner = Spinner("weather", text=text)
        with Live(spinner, refresh_per_second=3, console=Console(stderr=True)):
            while self.status in ("running", "scheduled", "unknown"):
                time.sleep(dt)
                n_itt += 1
                if n_itt > max_itt:
                    spinner.update(text=text + "ouch")
                    raise ValueError("Plugin did not finish")
            spinner.update(text=text + "ok")

    def get_result_paths(
        self,
        dtype: Literal["data", "plot"] = "data",
        glob_pattern: str = "*.nc",
    ) -> List[Path]:
        """Get all created paths of a certain data type.

        This method allows you to query all output files of the plugin run.
        You can either search for data files or plotted output.

        Parameters
        ----------
        dtype: str
            The data type of the returned paths. This should be either
            data or plot.
        glob_pattern: str, default: *.nc
            Refine the output by filtering the returned files by the given
            glob pattern. By default only netCDF files ("*.nc") are added
            to the list.

        Returns
        -------
        List[Path]: A list of paths matching the search constrains.

        Example
        -------
        We are going to use a plugin called ``dummypluginfolders``
        which creates plots and netCDF files. In this example we want to
        open all netCDF files (``dtype = 'data'``) that match the filename
        constraint ``*data.nc``.


        .. execute_code::

            import freva
            import xarray as xr
            res = freva.run_plugin("dummypluginfolders", variable="pr")
            dset = xr.open_mfdataset(
                res.get_result_paths(dtype="data", glob_pattern="*data.nc")
            )
            print(dset.attrs["variable"])


        """
        return [
            Path(path)
            for (path, metadata) in self._hist.get("result", {}).items()
            if metadata.get("type") == dtype and fnmatch(path, glob_pattern)
        ]


class config:
    """Override the default or set the freva system configuration file.

    With the help of this class you can not only (temporarily) override
    the default configuration file and use a configuration from another
    project, but you can also set a path to a configuration file if no
    configuration file has been set. Additionally you can set any plugin
    paths that are not part of the configuration file.

    Parameters
    ----------
    config_file: str | pathlib.Path, default: None
        Path to the (new) configuration file.
    plugin_path: str | List[str], default: None
        New plugins that should be used, use a list of paths if you want
        export multiple plugins.

    Examples
    --------
    Temporarily override the existing configuration file and use a new one.
    You can use a context manager to temporally use a different configuration
    and switch back later.

    ::

        import freva
        with freva.config("/work/freva/evaluation_system.conf"):
            freva.run_plugin("plugin_from_another_project")

    If you do not want to switch to another configuration only
    temporarily, but want to use it permanently, you can use
    :py:class:`freva.config` without a context manager:
    a context manager:

    ::

        import freva
        freva.config("/work/freva/evaluation_system.conf")
        files = sorted(freva.databrowser(project="user-1234", experiment="extremes"))

    Import a new user defined plugin, for example if you have created a plugin
    called ``MyPlugin`` that is located in ``~/freva/myplugin/plugin.py``
    you would set to ``plugin_path='~/freva/my_plugin,plugin_module'``.

    ::

        import freva
        freva.config(plugin_path="~/freva/my_plugin,plugin_module")
        freva.run_plugin('MyPlugin", variable1=1, variable2="a")

    In the same fashion you can set multiple plugin paths:

    ::
        import freva
        freva.config(plugin_path=["~/freva/my_plugin1,plugin_module_b"],
                                  "~/ freva/my_plugin2,plugin_module_b"])

    """

    _original_config_env = os.environ.get(
        "EVALUATION_SYSTEM_CONFIG_FILE",
    )
    _original_plugin_env = os.environ.get("EVALUATION_SYSTEM_PLUGINS")
    db_reloaded: List[bool] = [False]

    def __init__(
        self,
        config_file: Optional[Union[str, Path]] = None,
        plugin_path: Optional[Union[str, List[str]]] = None,
    ) -> None:
        plugin_path = plugin_path or []
        if isinstance(plugin_path, str):
            plugin_path = [plugin_path]
        if config_file:
            config_file = Path(config_file).expanduser().absolute()
            os.environ["EVALUATION_SYSTEM_CONFIG_FILE"] = str(config_file)
        try:
            if django_settings.DATABASES:
                self.db_reloaded[0] = True
            db_settings.reconfigure_django(config_file)
        except (ImproperlyConfigured, AttributeError):  # prgama: no cover
            pass  # prgama: no cover
        assert isinstance(db_settings.SETTINGS, dict)
        if plugin_path or config_file:
            pm.reload_plugins(plugin_path=plugin_path)

    def __enter__(self) -> "config":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        os.environ.pop("EVALUATION_SYSTEM_CONFIG_FILE")
        if self._original_config_env:
            os.environ["EVALUATION_SYSTEM_CONFIG_FILE"] = self._original_config_env
        db_settings.reconfigure_django(self._original_config_env)
        try:
            pm.reload_plugins()
        except Exception:
            pass
        self.db_reloaded[0] = False



"""
Created on 30.05.2016

@author: Sebastian Illing
"""

import json
import shlex
from pathlib import Path

from evaluation_system.tests import run_cli


def test_show_facet(capsys, dummy_config):
    run_cli("esgf --show-facet=project,product -d")
    res = capsys.readouterr().out
    assert "[product]" in res
    assert "[project]" in res


def test_query(capsys, search_dict, dummy_config):
    from freva.cli.esgf import main as run

    run_cli("esgf --debug project=TEST --query=project,product --opendap")
    res = capsys.readouterr().out
    res = json.loads(res)[0]
    assert "project" in res.keys()
    run_cli("esgf -d project=TEST limit=1 --query=project --gridftp")
    res = capsys.readouterr().out
    assert "['TEST']" in res
    run(shlex.split("project=TEST --query=project --gridftp"))
    res = capsys.readouterr().out
    res = [r for r in res.split("\n") if r.strip()]
    num_res = len(res)
    run_cli("esgf project=TEST offset=1 --query=project --gridftp")
    res = capsys.readouterr().out
    res = [r for r in res.split("\n") if r.strip()]
    assert num_res == len(res) + 1
    run(["project=TESTs"])
    assert "\n" in capsys.readouterr().out
    run(shlex.split("project=TEST --datasets"))
    assert "- version:" in capsys.readouterr().out
    run(["--show-facet=blabla"])
    assert not capsys.readouterr().out


def test_freva_esgf_method(dummy_config):
    from freva import esgf_browser, esgf_download, esgf_facets, esgf_query

    result_to_be = [
        "output1/MPI-M/MPI-ESM-LR/historical/day/atmos/day/r1i1p1/v20111006/"
        "tas/tas_day_MPI-ESM-LR_historical_r1i1p1_18500101-18591231.nc",
        "output1/MPI-M/MPI-ESM-LR/historical/day/atmos/day/r1i1p1/v20111006/"
        "tas/tas_day_MPI-ESM-LR_historical_r1i1p1_18600101-18691231.nc",
    ]
    res = " ".join(
        esgf_browser(
            project="CMIP5",
            experiment="historical",
            variable="tas",
            institute="MPI-M",
            time_frequency="day",
        )
    )
    for f in result_to_be:
        assert f in res
    result_to_be = [
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_201501010130-202412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_202501010130-203412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_203501010130-204412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_204501010130-205412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_205501010130-206412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_206501010130-207412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_207501010130-208412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_208501010130-209412312230.nc.html",
        "http://esgf3.dkrz.de/thredds/dodsC/cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_209501010130-210012312230.nc.html",
    ]
    res = " ".join(
        esgf_browser(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr",
            variable="uas",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
            opendap=True,
        )
    )
    for f in result_to_be:
        assert f in res
    result_to_be = [
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_201501010130-202412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_202501010130-203412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_203501010130-204412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_204501010130-205412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_205501010130-206412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_206501010130-207412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_207501010130-208412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_208501010130-209412312230.nc",
        "gsiftp://esgf3.dkrz.de:2811//cmip6/ScenarioMIP/CNRM-CERFACS/CNRM-CM6-1/ssp585/r1i1p1f2/E3hr/uas/gr/v20190219/uas_E3hr_CNRM-CM6-1_ssp585_r1i1p1f2_gr_209501010130-210012312230.nc",
    ]
    res = " ".join(
        esgf_browser(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr",
            variable="uas",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
            gridftp=True,
        )
    )
    for f in result_to_be:
        assert f in res
    result_to_be = {
        "version": "20190219",
        "activity_id": ["ScenarioMIP"],
        "master_id": "CMIP6.ScenarioMIP.CNRM-CERFACS.CNRM-CM6-1.ssp585.r1i1p1f2.E3hr.vas.gr",
        "mip_era": ["CMIP6"],
        "product": ["model-output"],
        "source_id": ["CNRM-CM6-1"],
        "url": [
            "http://esgf3.dkrz.de/thredds/catalog/esgcet/1586/CMIP6.ScenarioMIP.CNRM-CERFACS.CNRM-CM6-1.ssp585.r1i1p1f2.E3hr.vas.gr.v20190219.xml#CMIP6.ScenarioMIP.CNRM-CERFACS.CNRM-CM6-1.ssp585.r1i1p1f2.E3hr.vas.gr.v20190219|application/xml+thredds|THREDDS",
            "http://esgf3.dkrz.de/las/getUI.do?catid=649F6BBCDD98B0633CF0042D119CC69D_ns_CMIP6.ScenarioMIP.CNRM-CERFACS.CNRM-CM6-1.ssp585.r1i1p1f2.E3hr.vas.gr.v20190219|application/las|LAS",
        ],
        "variable": ["vas"],
        "score": 1.0,
    }
    res = esgf_query(
        mip_era="CMIP6",
        activity_id="ScenarioMIP",
        source_id="CNRM-CM6-1",
        institution_id="CNRM-CERFACS",
        experiment_id="ssp585",
        frequency="3hr",
        variant_label="r1i1p1f2",
        distrib=False,
        latest=True,
        query=[
            "url",
            "master_id",
            "distribution",
            "mip_era",
            "activity_id",
            "source_id",
            "variable",
            "product",
            "version",
        ],
    )
    for key, value in result_to_be.items():
        if key in res[:1]:
            assert res[:1][key] == result_to_be[key]
    res = esgf_facets(show_facet="product")
    res = res["product"]["MRE2reanalysis"]
    assert res == 6
    fn = Path("/tmp/file_script.sh")
    res = esgf_download(
        project="CMIP5",
        model="MPI-ESM-LR",
        experiment="decadal2001",
        variable="tas",
        distrib=False,
        download_script=fn,
    )
    assert fn.is_file()
    assert res == f"Download script successfully saved to {fn}"
    assert oct(fn.stat().st_mode)[-3:] == "755"
    res = esgf_download(
        project="CMIP5",
        model="MPI-ESM-LR",
        experiment="decadal2001",
        variable="tas",
        distrib=False,
    )
    fn = Path(res.split()[-1])
    assert fn.is_file()
    assert res == f"Download script successfully saved to {fn}"
    assert oct(fn.stat().st_mode)[-3:] == "755"


def test_find_files(capsys, search_dict, dummy_config):
    result_to_be = [
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/tas/1/"
        "tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/v20120529/"
        "tas/tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/v20120529/"
        "tas/tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
        "output1/MPI-M/MPI-ESM-LR/decadal2000/mon/atmos/Amon/r1i1p1/v20120529/"
        "tas/tas_Amon_MPI-ESM-LR_decadal2000_r1i1p1_200101-201012.nc",
    ]

    run_cli("esgf --show-facet=blabla")
    assert not capsys.readouterr().out
    run_cli(["esgf"] + [f"{key}={val}" for key, val in search_dict.items()])
    res = capsys.readouterr().out
    for f in result_to_be:
        assert f in res
    run_cli(
        ["esgf"] + ["--datasets"] + [f"{key}={val}" for key, val in search_dict.items()]
    )
    res = capsys.readouterr().out
    assert (
        "cmip5.output1.MPI-M.MPI-ESM-LR.decadal2000.mon.atmos.Amon.r1i1p1 - version: 20120529"
        in res
    )


def test_download_script(capsys, search_dict, tmp_dir, dummy_config):
    fn = tmp_dir / "download_test_script.sh"
    run_cli(
        ["esgf"]
        + ["%s=%s" % (key, val) for key, val in search_dict.items()]
        + [f"--download-script={fn}"]
    )
    res = capsys.readouterr().out
    assert fn.is_file()
    assert res == f"Download script successfully saved to {fn}\n"
    assert oct(fn.stat().st_mode)[-3:] == "755"



import os
import shutil
import tempfile
import time
from pathlib import Path
from subprocess import PIPE, run

from netCDF4 import Dataset as nc
from PIL import Image

from evaluation_system.api.parameters import (
    Directory,
    Float,
    InputDirectory,
    Integer,
    ParameterDictionary,
    String,
)
from evaluation_system.api.plugin import PluginAbstract
from evaluation_system.model.db import UserDB
from evaluation_system.model.user import User


class DummyPluginFolders(PluginAbstract):
    """Stub class for implementing the abstract one"""

    __short_description__ = "A dummy plugin with outputdir folder"
    __long_description__ = ""
    __tags__ = ["foo"]
    __version__ = (0, 1, 2)
    __category__ = "statistical"
    __name__ = "DummyPluginFolders"
    __parameters__ = ParameterDictionary(
        String(name="variable", default="tas", help="An input variable"),
        Directory(
            name="outputdir",
            default="$USER_OUTPUT_DIR/$SYSTEM_DATETIME",
            mandatory=False,
            help="The default output directory",
        ),
    )
    _runs: list = []
    _template = "${number} - $something - $other"
    # tool_developer = {"name": "DummyUser", "email": "data@dkrz.de"} it needs to be string or None
    tool_developer = "DummyUser"

    def run_tool(self, config_dict=None):
        DummyPluginFolders._runs.append(config_dict)
        out_dir = Path(config_dict["outputdir"]).absolute().expanduser()
        out_dir.mkdir(exist_ok=True, parents=True)
        image = Image.new("RGB", (300, 200), "white")
        image.save(str(out_dir / "plot.png"))
        with nc(str(out_dir / "data.nc"), "w") as dataset:
            dataset.variable = config_dict["variable"]
        print(f"Processing output in {config_dict['outputdir']}")
        return self.prepare_output(config_dict["outputdir"])



"""Module to get information on and run Freva plugins.

To make use of any of the methods a Freva plugin has to be set up. Either
by you, the Freva admins or your colleagues. If you want to create a plugin
pleas refer to the :class:`evaluation_system.api.plugin` section.
"""

from __future__ import annotations

import abc
import json
import logging
import textwrap
import time
import warnings
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, Literal, NamedTuple, Optional, Tuple, Union

import appdirs
import lazy_import
import rich.layout
import rich.panel
import rich.table

import freva
from evaluation_system.misc import logger

from .utils import PluginStatus, handled_exception, is_jupyter, meta_type

django = lazy_import.lazy_module("django")
pm = lazy_import.lazy_module("evaluation_system.api.plugin_manager")
user = lazy_import.lazy_module("evaluation_system.model.user")
config = lazy_import.lazy_module("evaluation_system.misc.config")
utils = lazy_import.lazy_module("evaluation_system.misc.utils")
PluginNotFoundError = lazy_import.lazy_callable(
    "evaluation_system.misc.exceptions.PluginNotFoundError"
)


CACHE_FILE = Path(appdirs.user_cache_dir()) / "freva" / "plugins.json"

PluginInfo = NamedTuple(
    "PluginInfo",
    [
        ("name", str),
        ("description", str),
        ("parameters", List[Tuple[str, str, Any]]),
    ],
)

__all__ = ["run_plugin", "list_plugins", "plugin_doc"]


class HelpStr:
    """A helper class that makes the string output prettier."""

    @abc.abstractmethod
    def __str__(self) -> str:
        """String conversion of the class."""

    @abc.abstractmethod
    def __repr__(self) -> str:
        """String representation of the class."""

    @abc.abstractmethod
    def _repr_html_(self) -> str:
        """Jupyter notebook representation of the class."""

    @abc.abstractmethod
    def __rich__(self) -> rich.table.Table:
        """Rich implementation of the class."""


def _write_plugin_cache() -> dict[str, dict[str, Any]]:
    """Write all plugins to a file."""
    out: dict[str, dict[str, Any]] = {}
    for plugin in pm.get_plugins().keys():
        pm_instance = pm.get_plugin_instance(plugin)
        out[plugin] = {}
        out[plugin]["description"] = pm_instance.__short_description__
        out[plugin]["parameters"] = []
        for key, param in pm_instance.__parameters__._params.items():
            if param.mandatory:
                desc = f"{param.help} (mandatory)"
            else:
                desc = f"{param.help} (default: {param.default})"
            out[plugin]["parameters"].append((key, desc, param.default))

    CACHE_FILE.parent.mkdir(exist_ok=True, parents=True)
    with CACHE_FILE.open("w") as f_obj:
        json.dump(out, f_obj)
    return out


@handled_exception
def read_plugin_cache(max_mtime: int = 180) -> list[PluginInfo]:
    """Cache plugin list."""
    CACHE_FILE.parent.mkdir(exist_ok=True, parents=True)
    CACHE_FILE.touch(exist_ok=True)
    plugin_cache = []
    if time.time() - CACHE_FILE.stat().st_mtime > max_mtime:
        cache = _write_plugin_cache()
    else:
        try:
            with CACHE_FILE.open("r") as f_obj:
                cache = json.load(f_obj)
        except:
            cache = _write_plugin_cache()
    if not cache:
        cache = _write_plugin_cache()
    for plugin, plugin_data in cache.items():
        plugin_cache.append(PluginInfo(name=plugin, **plugin_data))
    return plugin_cache


@handled_exception
def plugin_doc(tool_name: Optional[str]) -> HelpStr:
    """Display the documentation of a given plugin.

    Parameters
    ----------
    tool_name:
        The name of the tool that should be documented.

    Returns
    -------
    str:
        plugin help string.

    Raises
    ------
    PluginNotFoundError:
        if the plugin name does not exist.

    Example
    -------

    .. execute_code::

        import freva
        import rich
        rich.print(freva.plugin_doc("animator"))


    """
    tool_name = (tool_name or "").lower()
    _write_plugin_cache()
    _check_if_plugin_exists(tool_name)

    class HelpCls(HelpStr):
        def __init__(self, tool_name: str) -> None:
            self._plugin = pm.get_plugin_instance(tool_name)
            self._help = (
                self._plugin.__long_description__.strip()
                or self._plugin.__short_description__.strip()
            )
            self._version = ".".join([str(i) for i in self._plugin.__version__])
            self._name = self._plugin.__class__.__name__

        def __str__(self) -> str:
            return "{} (v{}): {}\n{}".format(
                self._name,
                self._version,
                self._help,
                self._plugin.__parameters__.get_help(),
            )

        def __repr__(self) -> str:
            return "{} (v{}): {}\n{}".format(
                self._plugin.__class__.__name__,
                self._version,
                self._help,
                self._plugin.__parameters__.get_help(),
            )

        def _repr_html_(self) -> str:
            return "<b>{}</b> (v{}): {}</p>{}".format(
                self._plugin.__class__.__name__,
                self._version,
                self._help.replace("\n", "<br>"),
                self._plugin.__parameters__.get_help(notebook=True),
            )

        def __rich__(self) -> rich.table.Table:
            help_text = rich.table.Text(
                self._help,
                overflow="fold",
                style="normal",
                justify="left",
            )
            title_text = rich.table.Text(
                overflow="fold", style="normal", justify="left"
            )
            title_text.append(
                self._plugin.__class__.__name__,
                style="bold",
            )
            title_text.append(" (")
            title_text.append(f"v{self._version}", style="italic")
            title_text.append("): ")
            title_text += help_text
            table = rich.table.Table(
                highlight=True,
                title=title_text,
                title_style="normal",
            )
            table.add_column("Option")
            table.add_column("Description")
            for key, param in self._plugin.__parameters__._params.items():
                param_str = param.format()
                var_name = key
                if param.mandatory:
                    var_name = f"[red][b]{var_name}[/b][/red]"
                param_desc = rich.table.Text(
                    f"{param.help} (default: {param_str})",
                    overflow="fold",
                )
                table.add_row(var_name, param_desc)
            return table

    return HelpCls(tool_name)


@handled_exception
def list_plugins() -> list[str]:
    """Get the plugins that are available on the system.

    Returns
    --------
    list[str]:
            List of available Freva plugins.

    Example
    -------

    .. execute_code::

        import freva
        import rich
        rich.print(freva.list_plugins())
    """
    _write_plugin_cache()
    return list([k.lower() for k in pm.get_plugins().keys()])


@handled_exception
def get_tools_list() -> HelpStr:
    """Get a list of plugins with their short description.

    Returns
    -------
    str:
        String representation of all available plugins.

    Example
    -------

    .. execute_code::

        import freva
        import rich
        rich.print(freva.get_tools_list())

    """
    _write_plugin_cache()

    class HelpCls(HelpStr):
        def __init__(self) -> None:
            env = utils.get_console_size()
            # we just have to show the list and stop processing
            name_width = 0
            self.plugins = pm.get_plugins()
            for key in self.plugins:
                name_width = max(name_width, len(key))
            self.offset = name_width + 2
            self.column_width = env["columns"] - self.offset
            self.result = self.constructor()

        def constructor(self) -> Dict[str, List[str]]:
            """Construct the things that should be displayed."""
            results = {}
            for _, plugin in sorted(self.plugins.items()):
                lines = textwrap.wrap(
                    str(plugin.description),
                    self.column_width,
                )
                if not lines:
                    lines = ["No description."]
                results[str(plugin.name)] = [lines[0]]
                if len(lines) > 1:
                    # multi-line
                    results[str(plugin.name)] += [
                        f"{' '*(len(plugin.name)+2)}{line}" for line in lines[1:]
                    ]
            return results

        def __rich__(self) -> rich.table.Table:
            table = rich.table.Table(highlight=True)
            table.add_column("Tool")
            table.add_column("Description")
            for plugin, desc in self.result.items():
                text = " ".join(" ".join(desc).split())
                table.add_row(plugin, rich.table.Text(text, overflow="fold"))
            return table

        def __str__(self) -> str:
            results = []
            for plugin, desc in self.result.items():
                results.append(f"{plugin}: " + "\n".join(desc))
            return "\n".join(results)

        def __repr__(self) -> str:
            return self.__str__()

        def _repr_html_(self) -> str:
            result = ["<table>"]
            for _, plugin in sorted(self.plugins.items()):
                result.append(
                    (
                        '<tr><td style="text-align: left;"><b>{}</b></td>'
                        '<td style="text-align: left;">{}</td></tr>'
                    ).format(plugin.name, plugin.description or "No description.")
                )
            result.append("</table>")
            return "".join(result)

    return HelpCls()


@handled_exception
def _check_if_plugin_exists(tool_name: Optional[str]) -> None:
    """Check if a given plugin name is part of the plugin stack."""
    if tool_name in pm.get_plugins():
        return
    if not tool_name:
        error = "Available tools are:\n"
        tool_list = "\n".join(list_plugins())
    else:
        tool_list = "\n".join(utils.find_similar_words(tool_name, list_plugins()))
        error = f"{tool_name} plugin not found, did you mean:\n"
    raise PluginNotFoundError(f"\n{error}{tool_list}")


def _get_tool_dict(
    tool_name: str, **options: Union[str, float, int, bool]
) -> Dict[str, Any]:
    options_str, tool_dict = [], {}
    for k, v in options.items():
        options_str.append(f"{k}={v}")
    tool_dict = pm.parse_arguments(tool_name, options_str)
    if logger.level == logging.DEBUG:
        tool_dict["debug"] = True
    return tool_dict


@handled_exception
def plugin_info(
    tool_name: str,
    what: Literal["repository", "config"] = "repository",
    **options: Union[str, float, int, bool],
) -> str:
    """Get additional information on a specific plugin.

    This is a utility function that lets you inspect additional information
    of a specific plugin. You can either get information on the plugin
    repository or get the current plugin configuration if you want to save
    the configuration to a file and use this file instead of calling the
    :py:meth:`freva.run_plugin` with the same options all over again.

    Parameters
    ----------
    tool_name: str
        The name of the plugin.
    what: str, default: repo_version
        What information should be returned. This can either be ``repo_version``
        for getting information on the tool repository or ``config`` for
        creating a string representing the current tool configuration that
        can then be saved to a file.
    **options:
        Parameters that should be passed to the tool.

    Returns
    -------
    str: The information that was requested.


    Example
    -------

    Get the repository and the last commit hash of a specific plugin:

    .. execute_code::

        import freva
        print(freva.plugin_info("animator", "repository"))

    Get the configuration for a plugin so it can be saved to a config file.

    .. execute_code::

        from tempfile import NamedTemporaryFile
        import freva
        config = freva.plugin_info("animator", what="config", variable="pr")
        print(config)
        with NamedTemporaryFile(suffix=".conf") as tf:
            with open(tf.name, "w") as f_obj:
                f_obj.write(config)

    """
    choices = ("repository", "config")
    if what not in choices:
        raise ValueError(f"what argument must be one of {', '.join(choices)}")
    if what == "repository":
        (repos, version) = pm.get_plugin_version(tool_name)
        return f"Repository and version of :{tool_name}\n{repos}\n{version}"
    tool_dict = _get_tool_dict(tool_name, **options)
    with NamedTemporaryFile() as tf:
        pm.write_setup(tool_name, tool_dict, config_file=tf.name)
        return Path(tf.name).read_text()


@handled_exception
def run_plugin(
    tool_name: str,
    *,
    save: bool = False,
    save_config: Optional[Union[str, Path]] = None,
    show_config: bool = False,
    scheduled_id: Optional[int] = None,
    unique_output: bool = True,
    batchmode: bool = False,
    caption: str = "",
    tag: Optional[str] = None,
    **options: Union[str, float, int, bool],
) -> PluginStatus:
    """Apply an available data analysis plugin.

    Parameters
    ----------
    tool_name:
        The name of the plugin that is to be applied.
    caption:
        Set a caption for the results.
    save:
        Save the plugin configuration to default destination.
    save_config:
        Save the plugin configuration.
    scheduled_id:
        Run a scheduled job from database
    batchmode:
        Create a Batch job and submit it to the scheduling system.
    unique_output:
        Append a Freva run id to the output/cache folder(s).
    tag:
       Use git commit hash to specify a specific version of this tool.

    Returns
    -------
    tuple:
        Return code, and the return value of the plugin.


    Example
    -------

    Run a plugin in the foreground.

    .. execute_code::

        import freva
        res = freva.run_plugin("animator", variable="pr", project="obs*")
        output = res.get_result_paths("plot", "*.*") # Check the plot output

    Run a plugin in the background. You can interact with the plugin using
    the ``.wait`` method of the :py:class:``freva.PluginStatus`` class.

    .. execute_code::

        import freva
        res = freva.run_plugin("animator",
                               variable="pr",
                               project="observations",
                               batchmode=True)
        res.wait() # Wait until the plugin has finished

    """
    tool_name = tool_name.lower()
    _check_if_plugin_exists(tool_name)
    if save_config:
        save_config = str(Path(save_config).expanduser().absolute())
    tool_dict: Dict[str, Any] = {}
    if scheduled_id is None:
        tool_dict = _get_tool_dict(tool_name, **options)
    if logger.level == logging.DEBUG:
        tool_dict["debug"] = True
    extra_scheduler_options = tool_dict.pop("extra_scheduler_options", "")
    if caption:
        caption = caption.strip()
    if save_config or save:
        save_in = pm.write_setup(tool_name, tool_dict, config_file=save_config)
        logger.info("Configuration file saved in %s", save_in)
    if scheduled_id:
        logger.info(
            "Running %s as scheduled in history with ID %i",
            tool_name,
            scheduled_id,
        )
        tool_id, result = pm.run_tool(
            tool_name, scheduled_id=scheduled_id, unique_output=unique_output
        )
        return PluginStatus(tool_id)
    extra_options: list[str] = [
        opt.strip() for opt in extra_scheduler_options.split(",") if opt.strip()
    ]
    # now run the tool
    result, tool_id = None, -1
    (error, warning) = pm.get_error_warning(tool_name)
    if warning:
        logger.warning(warning)
    if error:
        logger.error(error)
    logger.debug("Running %s with configuration: %s", tool_name, tool_dict)
    if not error:
        # we check if the user is external and activate batchmode
        django_user = django.contrib.auth.models.User.objects.get(
            username=user.User().getName()
        )
        if django_user.groups.filter(
            name=config.get("external_group", "noexternalgroupset")
        ).exists():
            batchmode = True
        if batchmode:
            if freva.config.db_reloaded[0]:
                raise ValueError(
                    "batchmode and overriding the config is not implemented"
                )
            [scheduled_id, job_file] = pm.schedule_tool(
                tool_name,
                config_dict=tool_dict,
                user=user.User(),
                caption=caption,
                extra_options=extra_options,
                unique_output=unique_output,
            )
            print(f"Scheduled job with history id: {scheduled_id}")
            print("You can view the job's status with the command squeue")
            print("Your job's progress will be shown with the command")
            print(f"tail -f {job_file}")
            return PluginStatus(scheduled_id or 0)
        if freva.config.db_reloaded[0]:
            warnings.warn(
                "History will not be available in currently active system but"
                " in previously active system.",
                UserWarning,
            )
        tool_id, result = pm.run_tool(
            tool_name,
            config_dict=tool_dict,
            caption=caption,
            unique_output=unique_output,
        )

        # repeat the warning at the end of the run
        # for readability don't show the warning in debug mode
        if warning:
            logger.warning(warning)
    logger.debug("Arguments: %s", options)
    logger.debug("Current configuration:\n%s", json.dumps(tool_dict, indent=4))
    return PluginStatus(tool_id)



"""
.. moduleauthor:: Sebastian Illing

This modules encapsulates all access to databases.
"""

import json
import re
import socket
from datetime import datetime
from typing import cast

import pandas as pd
from django.contrib.auth.models import User
from django.db import transaction

import evaluation_system.model.history.models as hist
import evaluation_system.model.plugins.models as pin
import evaluation_system.settings.database
from evaluation_system.misc import config
from evaluation_system.misc import logger as log
from evaluation_system.misc.utils import metadict
from evaluation_system.model.history.models import Configuration

TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


def timestamp_to_string(datetime_obj):
    return datetime_obj.strftime(TIMESTAMP_FORMAT)


def timestamp_from_string(date_string):
    """Create a datetime object from a given datetime string."""
    if date_string is None:
        return
    try:
        return pd.Timestamp(date_string).to_pydatetime()
    except:
        raise ValueError("Can't parse a date from '%s'" % date_string)


# class HistoryResultEntry(object):
#     """
#     This class encapsulates the access to the results.
#     """
#     def __init__(self, row):
#         self.id = row[0]
#         self.history_id_id = row[1]
#         self.output_file = row[2]
#         self.preview_file = row[3]
#         self.filetype = row[4]
#
#
# class HistoryTagEntry(object):
#     """
#     This class encapsulates the access to the HistoryTag entries.
#     """
#     def __init__(self, row):
#         self.id = row[0]
#         self.history_id_id = row[1]
#         self.type = row[2]
#         self.uid = row[3]
#         self.text = row[4]


class UserDB(object):
    """Encapsulates access to the local DB of a single user."""

    def __init__(self, user):
        self._user = user

    @transaction.atomic
    def storeHistory(
        self,
        tool,
        config_dict,
        uid,
        status,
        slurm_output=None,
        result=None,
        flag=None,
        version_details=None,
        caption=None,
    ):
        """Store an analysis run into the DB.

        :type tool: :class:`evaluation_system.api.plugin.pluginAbstract`
        :param tool: the plugin for which we are storing the information.
        :param config_dict: dictionary with the configuration used for this run,
        :param uid: the user id (useful in a global database)
        :param status: the process status
        :param result: dictionary with the results (created files)."""
        if result is None:
            result = {}
        if slurm_output is None:
            slurm_output = 0
        if flag is None:
            flag = 0
        if version_details is None:
            version_details = 1
        toolname = tool.__class__.__name__.lower()
        version = repr(tool.__version__)
        for key in config.exclude:
            config_dict.pop(key, "")
        newentry = hist.History(
            timestamp=datetime.now(),
            tool=toolname,
            version=version,
            configuration=json.dumps(config_dict),
            slurm_output=slurm_output,
            uid_id=uid,
            status=status,
            flag=flag,
            version_details_id=version_details,
        )
        # set caption
        if caption:
            newentry.caption = caption

        tool.__parameters__.synchronize(toolname)

        try:
            newentry.save()

            for p in tool.__parameters__._params.values():
                if p.name not in config.exclude:
                    name = p.name
                    param = Configuration(
                        history_id_id=newentry.id,
                        parameter_id_id=p.id,
                        value=json.dumps(config_dict[name]),
                        is_default=p.is_default,
                    )
                    param.save()
        except Exception as e:
            raise e

        return newentry.id

    def scheduleEntry(self, row_id, uid, slurmFileName, status=None):
        """Schedule a tool for a future application
        Parameter:
        ----------
        row_id:
            The index in the history table
        uid:
            the user id
        slurmFileName:
            The std out file belonging to the history entry
        status:
            Overwrite the default status (scheduled) with this status.
        """

        h = hist.History.objects.get(
            id=row_id,
            uid_id=uid,
            status=hist.History.processStatus.not_scheduled,
        )

        h.slurm_output = slurmFileName
        h.host = socket.gethostname().partition(".")[0]
        h.status = status or hist.History.processStatus.scheduled

        h.save()

    def store_batch_settings(
        self,
        row_id: int,
        job_script: str,
        workload_manager: str,
        job_id: int,
        output_file: str,
    ) -> None:
        """Store batch settings of the batch job for future interaction."""
        settings = hist.BatchSettings(
            history_id_id=row_id,
            job_script=job_script,
            workload_manager=workload_manager,
            job_id=job_id,
            output_file=output_file,
            host=socket.gethostname().partition(".")[0],
        )
        settings.save()

    class ExceptionStatusUpgrade(Exception):
        """
        Exception class for failing status upgrades
        """

        def __init__(self, msg="Status could not be upgraded"):
            super(UserDB.ExceptionStatusUpgrade, self).__init__(msg)

    def upgradeStatus(self, row_id, uid, status):
        """
        :param row_id: The index in the history table
        :param uid: the user id
        :param status: the new status
        After validation the status will be upgraded.
        """

        h = hist.History.objects.get(pk=row_id, uid_id=uid)

        if h.status < status:
            raise self.ExceptionStatusUpgrade("Tried to downgrade a status")

        h.status = status

        h.save()

    def changeFlag(self, row_id, uid, flag):
        """
        :param row_id: The index in the history table
        :param uid: the user id
        :param flag: the new flag
        After validation the status will be upgraded.
        """

        h = hist.History.objects.get(id=row_id, uid_id=uid)

        h.flag = flag

        h.save()

    def getHistory(
        self,
        tool_name=None,
        limit=-1,
        since=None,
        until=None,
        entry_ids=None,
        uid=None,
    ):
        """Returns the stored history (run analysis) for the given tool.

        :type tool_name: str
        :param tool_name: name of the tool for which the information will be gathered (if None, then everything is returned).
        :type limit: int
        :param limit: Amount of rows to be returned (if < 0, return all).
        :type since: datetime.datetime
        :param since: Return only items stored after this date
        :type until: datetime.datetime
        :param until: Return only  items stored before this date
        :param entry_ids: ([int] or int) id or list thereof to be selected
        :returns: ([:class:`HistoryEntry`]) list of entries that match the query.
        """
        filter_dict = {}

        if entry_ids is not None:
            if isinstance(entry_ids, int):
                entry_ids = [entry_ids]
            filter_dict["id__in"] = entry_ids

        if tool_name is not None:
            filter_dict["tool"] = tool_name

        if since is not None:
            filter_dict["timestamp__gte"] = since

        if until is not None:
            filter_dict["timestamp__lte"] = until

        if uid is not None:
            filter_dict["uid_id"] = uid

        o = hist.History.objects.filter(**filter_dict).order_by("-id")

        if limit > 0:
            o = o[:limit]

        return o

    def addHistoryTag(self, hrowid, tagType, text, uid=None):
        """
        :type hrowid: integer
        :param hrowid: the row id of the history entry where the results belong to
        :type tagType: integer
        :param tagType: the kind of tag
        :type: text: string
        :param: text: the text belonging to the tag
        :type: uid: string
        :param: uid: the user, default: None
        """

        h = hist.HistoryTag(history_id_id=hrowid, type=tagType, text=text)

        if uid is not None:
            h.uid_id = uid

        h.save()

    def updateHistoryTag(self, trowid, tagType=None, text=None, uid=None):
        """
        :type trowid: integer
        :param trowid: the row id of the tag
        :type tagType: integer
        :param tagType: the kind of tag
        :type: text: string
        :param: text: the text belonging to the tag
        :type: uid: string
        :param: uid: the user, default: None
        """

        h = hist.HistoryTag.objects.get(id=trowid, uid_id=uid)

        if tagType is not None:
            h.type = tagType

        if text is not None:
            h.text = text

        h.save()

    def store_output(self, rowid: int, results: metadict) -> None:
        """Store the output of a plugin run."""

        output = hist.Output(history_id_id=rowid, result=results)
        output.save()
        self.storeResults(
            rowid,
            cast(
                metadict,
                {
                    k: v
                    for (k, v) in results.items()
                    if v.get("type", "data") not in ("data",)
                },
            ),
        )

    def storeResults(self, rowid: int, results: metadict) -> None:
        """
        :type rowid: integer
        :param rowid: the row id of the history entry where the results belong to
        :type results: dict with entries {str : dict}
        :param results: meta-dictionary with meta-data dictionaries assigned to the file names.
        """
        reg_ex = None

        # regex to get the relative path
        preview_path = config.get(config.PREVIEW_PATH, None)
        expression = "(%s\\/*){1}(.*)" % re.escape(preview_path)

        # only try to create previews, when a preview path is given
        if preview_path:
            reg_ex = re.compile(expression)

        for file_name in results:
            metadata = results[file_name]

            type_name = metadata.get("type", "")
            type_number = hist.Result.Filetype.unknown

            preview_path = metadata.get("preview_path", "")
            preview_file = ""

            if preview_path and reg_ex is not None:
                # We store the relative path for previews only.
                # Which allows us to move the preview files to a different folder.
                preview_match = reg_ex.match(preview_path)
                if preview_match is not None:
                    preview_file = preview_match.group(2)

            if type_name == "plot":
                type_number = hist.Result.Filetype.plot
            elif type_name == "data":
                type_number = hist.Result.Filetype.data

            h = hist.Result(
                history_id_id=rowid,
                output_file=file_name,
                preview_file=preview_file,
                file_type=type_number,
            )

            h.save()

            result_id = h.pk
            self._storeResultTags(result_id, metadata)

    def _storeResultTags(self, result_id, metadata):
        """
        :type result_id: integer
        :param result_id: the id of the result entry where the tag belongs to
        :type metadata: dict with entries {str : dict}
        :param metadata: meta-dictionary with meta-data dictionaries assigned to the file names.
        """

        data_to_store = []

        # append new tags here
        caption = metadata.get("caption", None)

        if caption:
            data_to_store.append(
                hist.ResultTag(
                    result_id_id=result_id,
                    type=hist.ResultTag.flagType.caption,
                    text=caption,
                )
            )

        hist.ResultTag.objects.bulk_create(data_to_store)

    def getVersionId(
        self,
        toolname,
        version,
        repos_api,
        internal_version_api,
        repos_tool,
        internal_version_tool,
    ):
        repository = "%s;%s" % (repos_tool, repos_api)

        retval = None

        try:
            p = pin.Version.objects.filter(
                tool=toolname,
                version=version,
                internal_version_tool=internal_version_tool[:40],
                internal_version_api=internal_version_api[:40],
                repository=repository,
            )[0]

            retval = p.pk

        except (IndexError, pin.Version.DoesNotExist) as e:
            pass

        return retval

    def newVersion(
        self,
        toolname,
        version,
        repos_api,
        internal_version_api,
        repos_tool,
        internal_version_tool,
    ):
        repository = "%s;%s" % (repos_tool, repos_api)

        p = pin.Version(
            timestamp=datetime.now(),
            tool=toolname,
            version=version,
            internal_version_tool=internal_version_tool,
            internal_version_api=internal_version_api,
            repository=repository,
        )

        p.save()

        result_id = p.pk

        return result_id

    def getUserId(self, username):
        retval = 0

        try:
            u = User.objects.get(username=username)

            retval = u.pk
        except User.DoesNotExist:
            pass

        return retval

    def updateUserLogin(self, row_id, email=None):
        u = User.objects.get(id=row_id)

        u.last_login = datetime.now()

        if email is not None:
            u.email = email

        u.save()

    def createUser(
        self,
        username,
        email="-",
        first_name="",
        last_name="",
    ):
        timestamp = datetime.now()

        u = User(
            username=username,
            password="NoPasswd",
            date_joined=timestamp,
            last_login=timestamp,
            first_name=first_name,
            last_name=last_name,
            email=email,
            is_active=1,
            is_staff=0,
            is_superuser=0,
        )

        u.save()

    def create_user_crawl(self, crawl_dir, username):
        from evaluation_system.model.solr_models.models import UserCrawl

        crawl = UserCrawl(
            status="waiting",
            user_id=self.getUserId(username),
            path_to_crawl=crawl_dir,
        )
        crawl.save()
        return crawl.id



"""A Python module to access the apache solr databrowser."""

from __future__ import annotations

import json
import warnings
from pathlib import Path
from typing import Any, Iterator, Optional, Union, overload

import lazy_import
from typing_extensions import Literal

from evaluation_system.misc import logger

from .utils import handled_exception

SolrFindFiles = lazy_import.lazy_class("evaluation_system.model.solr.SolrFindFiles")


__all__ = ["databrowser", "search_facets", "count_values"]


def _proc_search_facets(
    time_select: Literal["flexible", "strict", "file"] = "flexible",
    **search_facets: str | list[str] | int,
) -> dict[str, str | list[str] | int]:
    """Correct the facet values if needed."""
    select_methods: dict[str, str] = {
        "strict": "Within",
        "flexible": "Intersects",
        "file": "Contains",
    }
    try:
        search_facets["time_select"] = select_methods[time_select]
    except KeyError as error:
        methods = ", ".join(select_methods.keys())
        raise ValueError(f"Time select method has to be one of {methods}") from error
    search_facets["time"] = search_facets.get("time", "")
    for key in ("file", "uri"):
        try:
            search_facets[key] = json.dumps(search_facets[key])
        except KeyError:
            pass
    return search_facets


@overload
def count_values(
    *,
    facet: Union[str, list[str]],
    time: str = "",
    time_select: Literal["strict", "flexible", "file"] = "flexible",
    multiversion: bool = False,
    **search_facets: str | list[str] | int,
) -> dict[str, dict[str, int]]: ...


@overload
def count_values(
    *,
    facet: Literal[None],
    time: str = "",
    time_select: Literal["strict", "flexible", "file"] = "flexible",
    multiversion: bool = False,
    **search_facets: str | list[str] | int,
) -> int: ...


@handled_exception
def count_values(
    *,
    time: str = "",
    time_select: Literal["strict", "flexible", "file"] = "flexible",
    multiversion: bool = False,
    facet: str | list[str] | None = None,
    **search_facets: str | list[str] | int,
) -> int | dict[str, dict[str, int]]:
    """Count the number of found objects in the databrowser.

    Parameters
    ----------
    time: str, default: ""
        Special search facet to refine/subset search results by time.
        This can be a string representation of a time range or a single
        time step. The time steps have to follow ISO-8601. Valid strings are
        ``%Y-%m-%dT%H:%M`` to ``%Y-%m-%dT%H:%M`` for time ranges and
        ``%Y-%m-%dT%H:%M``. **Note**: You don't have to give the full string
        format to subset time steps ``%Y``, ``%Y-%m`` etc are also valid.
    time_select: str, default: flexible
        Operator that specifies how the time period is selected. Choose from
        flexible (default), strict or file. ``strict`` returns only those files
        that have the *entire* time period covered. The time search ``2000 to
        2012`` will not select files containing data from 2010 to 2020 with
        the ``strict`` method. ``flexible`` will select those files as
        ``flexible`` returns those files that have either start or end period
        covered. ``file`` will only return files where the entire time
        period is contained within *one single* file.
    multiversion: bool, default: False
        Select all versions and not just the latest version (default).
    facet: Union[str, list[str]], default: None
        Count these these facets (attributes & values) instead of the number
        of total files. If None (default), the number of total files will
        be returned.
    **search_facets: str
        The facets to be applied in the data search. If not given
        the whole dataset will be queried.

    Returns
    -------
    int, dict[str, int]:
        Number of found objects, if the *facet* key is/are given then the
        a dictionary with the number of objects for each search facet/key
        is given.

    Example
    -------
    .. execute_code::

        import freva
        num_files = freva.count_values(experiment="cmorph")
        print(num_files)

    .. execute_code::

        import freva
        print(freva.count_values(facet="*"))

    """
    search_facets = _proc_search_facets(
        time_select=time_select, time=time, **search_facets
    )
    count_all = facet is None
    if isinstance(facet, str):
        facet = [facet]
    if facet in (["*"], ["all"]):
        facet = []
    facet = facet or []
    latest = not multiversion
    if "version" in search_facets and latest:
        # it makes no sense to look for a specific version just among the latest
        # the speedup is marginal and it might not be what the user expects
        logger.warning("Turning latest off when searching for a specific version.")
        latest = False
    core = {True: "latest", False: "files"}[latest]
    logger.debug("Searching dictionary: %s\n", search_facets)
    search_facets["facet.limit"] = search_facets.pop("facet_limit", -1)
    if count_all:
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=PendingDeprecationWarning)
            return (
                SolrFindFiles(core=core)._retrieve_metadata(**search_facets).num_objects
            )
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=PendingDeprecationWarning)
        results = SolrFindFiles(core=core)._facets(facet or None, **search_facets)
    out: dict[str, dict[str, int]] = {}
    for att in facet or results.keys():
        values = results[att]
        out[att] = {str(v): int(c) for v, c in zip(*[iter(values)] * 2)}
    return out


@handled_exception
def facet_search(
    *,
    time: str = "",
    time_select: Literal["strict", "flexible", "file"] = "flexible",
    multiversion: bool = False,
    facet: str | list[str] | None = None,
    **search_facets: str | list[str] | int,
) -> dict[str, list[str]]:
    """Search for data attributes (facets) in the databrowser.

    The method queries the databrowser for available search facets (keys)
    like model, experiment etc.

    Parameters
    ----------
    time: str
        Special search facet to refine/subset search results by time.
        This can be a string representation of a time range or a single
        time step. The time steps have to follow ISO-8601. Valid strings are
        ``%Y-%m-%dT%H:%M`` to ``%Y-%m-%dT%H:%M`` for time ranges and
        ``%Y-%m-%dT%H:%M``. **Note**: You don't have to give the full string
        format to subset time steps ``%Y``, ``%Y-%m`` etc are also valid.
    time_select: str, default: flexible
        Operator that specifies how the time period is selected. Choose from
        flexible (default), strict or file. ``strict`` returns only those files
        that have the *entire* time period covered. The time search ``2000 to
        2012`` will not select files containing data from 2010 to 2020 with
        the ``strict`` method. ``flexible`` will select those files as
        ``flexible`` returns those files that have either start or end period
        covered. ``file`` will only return files where the entire time
        period is contained within *one single* file.
    facet: Union[str, list[str]], default: None
        Retrieve information about these facets (attributes & values).
        If None given (default), information about all available facets is
        returned.
    multiversion: bool, default: False
        Select all versions and not just the latest version (default).
    **search_facets: str
        The facets to be applied in the data search. If not given
        the whole dataset will be queried.

    Returns
    -------
    dict[str, list[str]]:
        Dictionary with a list search facet values for each search facet key.


    Example
    -------

    .. execute_code::

        import freva
        all_facets = freva.facet_search(project='obs*')
        print(all_facets)
        spec_facets = freva.facet_search(project='obs*',
                                         facet=["time_frequency", "variable"])
        print(spec_facets)

    Get all models that have a given time step:

    .. execute_code::

        import freva
        model = list(freva.facet_search(project="obs*", time="2016-09-02T22:10"))
        print(model)

    Reverse search: retrieving meta data from a known file.

    .. execute_code::

        import os, freva
        file = "../.docker/data/observations/grid/CPC/CPC/cmorph/30min/atmos/30min/r1i1p1/v20210618/pr/pr_30min_CPC_cmorph_r1i1p1_201609020000-201609020030.nc"
        res = freva.facet_search(file=str(os.path.abspath(file)))
        print(res)

    """
    search_facets = _proc_search_facets(
        time_select=time_select, time=time, **search_facets
    )
    if isinstance(facet, str):
        facet = [facet]
    if facet in (["*"], ["all"]):
        facet = []
    facet = facet or []
    latest = not multiversion
    if "version" in search_facets and latest:
        # it makes no sense to look for a specific version just among the latest
        # the speedup is marginal and it might not be what the user expects
        logger.warning("Turning latest off when searching for a specific version.")
        latest = False
    core = {True: "latest", False: "files"}[latest]
    logger.debug("Searching dictionary: %s\n", search_facets)
    search_facets["facet.limit"] = search_facets.pop("facet_limit", -1)
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=PendingDeprecationWarning)
        results = SolrFindFiles(core=core)._facets(
            facets=facet or None, latest_version=False, **search_facets
        )
    return {f: v[::2] for f, v in results.items()}


@handled_exception
def databrowser(
    *,
    multiversion: bool = False,
    batch_size: int = 5000,
    uniq_key: Literal["file", "uri"] = "file",
    time: str = "",
    time_select: Literal["flexible", "strict", "file"] = "flexible",
    **search_facets: Union[str, list[str], int],
) -> Union[dict[str, dict[str, int]], dict[str, list[str]], Iterator[str], int]:
    """Find data in the system.

    You can either search for files or data facets (variable, model, ...)
    that are available. The query is of the form key=value. <value> might
    use *, ? as wildcards or any regular expression.

    Parameters
    ----------
    **search_facets: Union[str, Path, in, list[str]]
        The facets to be applied in the data search. If not given
        the whole dataset will be queried.
    time: str
        Special search facet to refine/subset search results by time.
        This can be a string representation of a time range or a single
        time step. The time steps have to follow ISO-8601. Valid strings are
        ``%Y-%m-%dT%H:%M`` to ``%Y-%m-%dT%H:%M`` for time ranges and
        ``%Y-%m-%dT%H:%M``. **Note**: You don't have to give the full string
        format to subset time steps ``%Y``, ``%Y-%m`` etc are also valid.
    time_select: str, default: flexible
        Operator that specifies how the time period is selected. Choose from
        flexible (default), strict or file. ``strict`` returns only those files
        that have the *entire* time period covered. The time search ``2000 to
        2012`` will not select files containing data from 2010 to 2020 with
        the ``strict`` method. ``flexible`` will select those files as
        ``flexible`` returns those files that have either start or end period
        covered. ``file`` will only return files where the entire time
        period is contained within *one single* file.
    uniq_key: str, default: file
        Chose if the solr search query should return paths to files or
        uris, uris will have the file path along with protocol of the storage
        system. Uris can be useful if the the search query result should be
        used libraries like fsspec.
    multiversion: bool, default: False
        Select all versions and not just the latest version (default).
    batch_size: int, default: 5000
        Size of the search query.

    Returns
    -------
    Iterator :
        If ``all_facets`` is False and ``facet`` is None an
        iterator with results.


    Example
    -------

    Search for files in the system:

    .. execute_code::

        import freva
        files = freva.databrowser(project='obs*', institute='cpc',
                                  time_frequency='??min',
                                  variable='pr')
        print(files)
        print(next(files))
        for file in files:
            print(file)
            break

    Search for files between a two given time steps:

    .. execute_code::

        import freva
        file_range = freva.databrowser(project="obs*", time="2016-09-02T22:15 to 2016-10")
        for file in file_range:
            print(file)

    The default method for selecting time periods is ``flexible``, which means
    all files are selected that cover at least start or end date. The
    ``strict`` method implies that the *entire* search time period has to be
    covered by the files. Using the ``strict`` method in the example above would
    only yield one file because the first file contains time steps prior to the
    start of the time period:

    .. execute_code::

        import freva
        file_range = freva.databrowser(project="obs*", time="2016-09-02T22:15 to 2016-10", time_select="strict")
        for file in file_range:
            print(file)

    In datasets with multiple versions only the `latest` version (i.e. `highest`
    version number) is returned by default. Querying a specific version from a
    multi versioned datasets requires the ``multiversion`` flag in combination with
    the ``version`` special facet:

    .. execute_code::

        import freva
        specific_version = list(freva.databrowser(project="reanalysis", version="v20200101", multiversion=True))
        print(specific_version)

    """
    core = {True: "latest", False: "files"}[not multiversion]
    search_facets = _proc_search_facets(
        time_select=time_select, time=time, **search_facets
    )
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=PendingDeprecationWarning)
        search_results = SolrFindFiles(core=core)._search(
            batch_size=batch_size,
            latest_version=not multiversion,
            uniq_key=uniq_key,
            **search_facets,
        )
    return search_results



import os

import mock
import pytest

SUBCOMMANDS = ("databrowser", "esgf", "user-data", "history", "plugin")


def test_main_help(capsys):
    from freva.cli import main as main_cli

    env = os.environ.copy()
    env.pop("EVALUATION_SYSTEM_CONFIG_FILE", "")
    with mock.patch.dict(os.environ, env, clear=True):
        with pytest.raises(SystemExit):
            main_cli(["-h"])
        doc_string = capsys.readouterr().out
        for subcommand in SUBCOMMANDS:
            assert subcommand in doc_string


def test_subcommand_help(capsys):
    from freva.cli import main as main_cli
    from freva.cli.databrowser import main as databrowser
    from freva.cli.esgf import main as esgf
    from freva.cli.history import main as history
    from freva.cli.plugin import main as plugin
    from freva.cli.user_data import main as user_data

    functions = (databrowser, esgf, user_data, history, plugin)
    env = os.environ.copy()
    env.pop("EVALUATION_SYSTEM_CONFIG_FILE", "")
    with mock.patch.dict(os.environ, env, clear=True):
        for subcommand, func in dict(zip(SUBCOMMANDS, functions)).items():
            with pytest.raises(SystemExit):
                func(["--help"])
            doc_string1 = " ".join(capsys.readouterr().out.replace(" [-V]", "").split())
            with pytest.raises(SystemExit):
                main_cli([subcommand, "--help"])
            doc_string2 = " ".join(
                capsys.readouterr()
                .out.replace(f"freva {subcommand}", f"freva-{subcommand}")
                .split()
            )
            for line in doc_string2.split("\n"):
                print(line)
                assert line.strip() in doc_string1
            with pytest.raises(SystemExit):
                main_cli([subcommand[:-1]])
            doc_string3 = capsys.readouterr().err
            assert subcommand in doc_string3



from __future__ import annotations

import argparse
import json
import sys
from typing import Any, Optional

import lazy_import
import rich

from evaluation_system import __version__
from evaluation_system.misc import logger

from .utils import BaseParser, standard_main

freva = lazy_import.lazy_module("freva")


class Cli(BaseParser):
    """Class that constructs the History Query Parser."""

    desc = "Read the plugin application history."

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the history sub arg. parser."""
        super().__init__(parser, "freva-history")
        self.parser.add_argument(
            "--full-text",
            default=False,
            action="store_true",
            help="Show the complete configuration.",
        )
        self.parser.add_argument(
            "--return-command",
            default=False,
            action="store_true",
            help=(
                "Show freva commands belonging to the history entries "
                "instead of the entries themselves."
            ),
        )
        self.parser.add_argument(
            "--limit",
            default=10,
            type=int,
            help="Limit the number of displayed entries to N.",
        )
        self.parser.add_argument(
            "--plugin",
            default=None,
            type=str,
            help="Display only entries of selected plugin.",
        )
        self.parser.add_argument(
            "--since",
            default=None,
            type=str,
            help="Retrieve entries older than date.",
        )
        self.parser.add_argument(
            "--until",
            default=None,
            type=str,
            help="Retrieve entries newer than date.",
        )
        self.parser.add_argument(
            "--entry-ids",
            default=None,
            type=str,
            nargs="+",
            help="Select entry id(s) (e.g. --entry-ids 1 --entry-ids 2 ).",
        )
        self.parser.add_argument(
            "--user-name",
            default=None,
            type=str,
            help=(
                "Show the history of other user (e.g. --user-name <username>)"
                " or all users (--user-name all or --user-name '*')."
            ),
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--json",
            "-j",
            help=(
                "Display a json representation of the output, this can be"
                "useful if you want to build shell based pipelines and want"
                "parse the output with help of `jq`."
            ),
            action="store_true",
            default=False,
        )

        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace,
        **kwargs: Any,
    ) -> None:
        """Call the databrowser command and print the results."""
        kwargs.pop("other_args", "")
        return_json = kwargs.pop("json", False)
        if return_json:
            kwargs["return_results"] = True
        try:
            if len(kwargs["entry_ids"]) == 1:
                kwargs["entry_ids"] = kwargs["entry_ids"][0].split(",")
        except TypeError:
            pass
        commands = freva.history(_return_dict=return_json, **kwargs)
        if not commands and return_json is False:
            return
        if args.return_command and return_json is False:
            result = "\n".join(commands)
        elif return_json is True:
            result = json.dumps(commands, indent=3)
        else:
            result = "\n".join(
                [c.__str__(compact=not args.full_text) for c in commands]
            )
        print(result)


def main(argv: Optional[list[str]] = None) -> None:
    standard_main(Cli, __version__, argv)



"""Module to print command argument line argument choices."""

import logging
import sys

import lazy_import

from evaluation_system.misc import logger

print_choices = lazy_import.lazy_function("freva.cli.utils.print_choices")


logger.setLevel(logging.ERROR)

try:
    print_choices(sys.argv[1:])
except KeyboardInterrupt:
    pass



"""
Created on 31.05.2016

@author: Sebastian Illing
"""

import datetime
import logging
import os
import shutil
import tempfile
import textwrap
import time
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from evaluation_system.api.plugin_manager import (  # _PluginStateHandle,
    CommandConfig,
    PluginMetadata,
)

# from evaluation_system.model.user import User

# from unittest.mock import MagicMock, patch


log = logging.getLogger(__name__)

# def test_update_plugin_state_in_db():
#     user_mock = MagicMock(spec=User)
#     plugin_state_handle = _PluginStateHandle(status=1, rowid=123, user=user_mock)
#     plugin_state_handle._update_plugin_state_in_db()
#     user_mock.getUserDB.return_value.upgradeStatus.assert_called_once_with(123, 'test_user', 1)


# @patch("evaluation_system.api.plugin_manager.atexit.register")
# @patch("evaluation_system.api.plugin_manager.signal.signal")
# def test_signal_handling_updates_state_and_quits(mock_signal, mock_atexit_register):
#     user_mock = MagicMock(spec=User)
#     plugin_state_handle = _PluginStateHandle(status=1, rowid=123, user=user_mock)
#     with patch.object(plugin_state_handle, "_update_plugin_state_in_db") as mock_update:
#         exception_caught = False
#         try:
#             plugin_state_handle._update_plugin_state_in_db_and_quit(None, None)
#         except KeyboardInterrupt:
#             exception_caught = True
#         assert exception_caught == True, "KeyboardInterrupt was not raised as expected"


def test_missing_plugin_directory_logs_warning(temp_user):
    import evaluation_system.api.plugin_manager as pm

    user_name = temp_user.getName()
    plugin_env_name = pm.PLUGIN_ENV + "_" + user_name
    non_existent_path = "/path/to/nonexistent/plugin"
    os.environ[plugin_env_name] = f"non_existent_module:{non_existent_path}"
    pm.reload_plugins(user_name=user_name)

    # Clean up by removing the environment variable to avoid side effects
    del os.environ[plugin_env_name]


def test_modules(dummy_settings):
    import evaluation_system.api.plugin_manager as pm

    pm.reload_plugins()
    pmod = pm.__plugin_modules_user__
    assert pmod is not None
    assert len(pmod) > 0


def test_get_plugins_user(temp_user):
    import evaluation_system.api.plugin_manager as pm

    get_plugins_user = pm.get_plugins_user()
    assert get_plugins_user == {
        str(temp_user.getName()): {
            "dummyplugin": PluginMetadata(
                name="DummyPlugin",
                plugin_class="DummyPlugin",
                plugin_module=str(os.getcwd())
                + "/src/evaluation_system/tests/mocks/dummy",
                description="A dummy plugin",
                user_exported=False,
                category="statistical",
                tags=["foo"],
            )
        }
    }


# @patch("evaluation_system.api.plugin_manager.importlib.util.spec_from_file_location")
# @patch("evaluation_system.api.plugin_manager.get_plugin_metadata")
# def test_get_plugin_instance_conditions(
#     mock_get_plugin_metadata, mock_spec_from_file_location
# ):
#     import evaluation_system.api.plugin_manager as pm

#     mock_plugin_metadata = MagicMock()
#     mock_plugin_metadata.plugin_class = "PluginClass"
#     mock_plugin_metadata.plugin_module = "plugin_module"
#     mock_get_plugin_metadata.return_value = mock_plugin_metadata

#     # First condition: spec is None
#     mock_spec_from_file_location.return_value = None
#     with pytest.raises(ImportError) as excinfo:
#         pm.get_plugin_instance("NonExistentPlugin")
#     assert "Could not import PluginClass from plugin_module" in str(excinfo.value)


def test_plugins(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm
    from evaluation_system.tests.mocks.dummy import DummyPlugin

    # force reload to be sure the dummy is loaded
    assert len(pm.get_plugins()) > 0
    assert "dummyplugin" in pm.get_plugins()
    dummy = pm.get_plugin_metadata("dummyplugin")
    assert dummy.description == DummyPlugin.__short_description__
    assert dummy.category == DummyPlugin.__category__
    assert dummy.tags == DummyPlugin.__tags__
    assert dummy.plugin_class == "DummyPlugin"
    os.environ[f"EVALUATION_SYSTEM_PLUGINS_{temp_user.getName()}"] = (
        f"{str(Path(__file__).parent / 'mocks')},dummy"
    )
    pm.reload_plugins(temp_user.getName())
    os.environ.pop(f"EVALUATION_SYSTEM_PLUGINS_{temp_user.getName()}")


def testDefaultPluginConfigStorage(temp_user):
    import evaluation_system.api.plugin_manager as pm

    pm.reload_plugins(temp_user.username)
    home = temp_user.getUserHome()
    assert os.path.isdir(home)
    conf_file = pm.write_setup("dummyplugin", user=temp_user)
    assert os.path.isfile(conf_file)


def test_plugin_config_storage(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm
    from evaluation_system.api.parameters import ValidationError

    home = temp_user.getUserHome()
    assert os.path.isdir(home)

    res = pm.get_plugin_instance("dummyplugin").setup_configuration(
        config_dict=dict(the_number=42)
    )
    assert res["something"] == "test"

    # write down this default
    conf_file = pm.write_setup(
        "dummyplugin", config_dict=dict(the_number=42), user=temp_user
    )

    assert os.path.isfile(conf_file)
    with open(conf_file, "r") as f:
        config = f.read()

    assert "\nsomething=test\n" in config

    with pytest.raises(ValidationError):
        pm.parse_arguments("dummyplugin", [])
    with pytest.raises(ValidationError):
        pm.parse_arguments("dummyplugin", [], user=temp_user)
    res = pm.parse_arguments("dummyplugin", [], use_user_defaults=True, user=temp_user)

    assert res == {
        "other": 1.4,
        "number": None,
        "the_number": 42,
        "something": "test",
        "input": None,
        "variable": "tas",
        "extra_scheduler_options": "",
    }
    assert res["something"] == "test"

    # now change the stored configuration
    config = config.replace("\nsomething=test\n", "\nsomething=super_test\n")
    with open(conf_file, "w") as f:
        f.write(config)
    res = pm.parse_arguments("dummyplugin", [], use_user_defaults=True, user=temp_user)
    assert res["something"] == "super_test"


def test_parse_arguments(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm

    home = temp_user.getUserHome()
    assert os.path.isdir(home)

    # direct parsing
    for args, result in [
        (
            "the_number=4",
            {
                "other": 1.3999999999999999,
                "the_number": 4,
                "something": "test",
                "variable": "tas",
                "extra_scheduler_options": "",
            },
        )
    ]:
        d = pm.parse_arguments("Dummyplugin", args.split(), user=temp_user)
        assert d == result

    # parsing requesting user default but without any
    for args, result in [
        (
            "the_number=4",
            {
                "other": 1.3999999999999999,
                "the_number": 4,
                "something": "test",
                "variable": "tas",
                "extra_scheduler_options": "",
            },
        )
    ]:
        d = pm.parse_arguments("Dummyplugin", args.split(), user=temp_user)
        assert d == result

    pm.write_setup("DummyPlugin", dict(number=7, the_number=42), temp_user)
    for args, result in [
        (
            "number=4",
            dict(
                number=4,
                the_number=42,
                something="test",
                other=1.4,
                input=None,
                variable="tas",
                extra_scheduler_options="",
            ),
        )
    ]:
        d = pm.parse_arguments(
            "Dummyplugin", args.split(), use_user_defaults=True, user=temp_user
        )
        assert d == result

    shutil.rmtree(home)


def test_parse_arguments_with_config_file(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm

    home = temp_user.getUserHome()
    assert os.path.isdir(home)

    with tempfile.NamedTemporaryFile(delete=False, mode="w") as tmp:
        tmp.write("[DummyPlugin]\nnumber=3\nthe_number=5")
        tmp_path = tmp.name

    # Parsing with a config_file
    d = pm.parse_arguments(
        "Dummyplugin", "number=4".split(), config_file=tmp_path, user=temp_user
    )
    assert d == {
        "number": 4,
        "the_number": 5,
        "something": "test",
        "other": 1.4,
        "input": None,
        "variable": "tas",
        "extra_scheduler_options": "",
    }

    # Cleanup
    os.remove(tmp_path)
    shutil.rmtree(home)


def test_write_setup(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm

    f = pm.write_setup(
        "DummyPlugin", dict(number="$the_number", the_number=42), temp_user
    )

    with open(f) as fp:
        num_line = [
            line for line in fp.read().splitlines() if line.startswith("number")
        ][0]
        assert num_line == "number=$the_number"


def test_get_history(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm
    import freva

    pm.write_setup("DummyPlugin", dict(the_number=777), temp_user)
    pm.run_tool("dummyplugin", user=temp_user)
    run_res = freva.run_plugin("dummyplugin", the_number=2, other=-5, batchmode=True)
    pm.unfollow_history_tag(run_res._id, user=temp_user)
    pm.run_tool("dummyplugin", scheduled_id=run_res._id, user=temp_user)
    run_res.kill()
    res = pm.get_history(user=temp_user)
    res = res[0]
    import re

    mo = re.search(
        "^([0-9]{1,})[)] ([^ ]{1,}) ([^ ]{1,}) ([^ ]{1,})", res.__str__(compact=False)
    )
    assert mo is not None
    g1 = mo.groups()
    assert all([g is not None for g in g1])
    mo = re.search("^([0-9]{1,})[)] ([^ ]{1,}) ([^ ]{1,})", res.__str__())
    g2 = mo.groups()
    assert all([g is not None for g in g2])
    assert g1[0] == g2[0]


def testDynamicPluginLoading(dummy_env, temp_user):
    import evaluation_system.api.plugin_manager as pm

    basic_plugin = textwrap.dedent(
        """
        from sys import modules
        plugin = modules['evaluation_system.api.plugin']
        parameters = modules['evaluation_system.api.parameters']

        class %s(plugin.PluginAbstract):
            __short_description__ = "Test"
            __version__ = (0,0,1)
            __category__ = "foo"
            __tags__ = ["bar"]
            __parameters__ =  parameters.ParameterDictionary(
                                            parameters.File(name="output", default="/tmp/file", help='output'),
                                            parameters.File(name="input", mandatory=True, help="some input"))

            def runTool(self, config_dict=None):
                print("%s", config_dict)
        """
    )

    with tempfile.TemporaryDirectory(prefix="dyn_plugin") as path1:
        os.makedirs(os.path.join(path1, "a/b"))
        with open(path1 + "/a/__init__.py", "w"):
            pass
        with open(path1 + "/a/blah.py", "w") as f:
            f.write(basic_plugin % tuple(["TestPlugin1"] * 2))

        with tempfile.TemporaryDirectory(prefix="dyn_plugin") as path2:
            os.makedirs(os.path.join(path2, "x/y/z"))
            with open(path2 + "/x/__init__.py", "w"):
                pass
            with open(path2 + "/x/foo.py", "w") as f:
                f.write(basic_plugin % tuple(["TestPlugin2"] * 2))

            os.environ[pm.PLUGIN_ENV] = "%s,%s:%s,%s" % (
                "~/../../../../../.." + path1 + "/a",
                "blah",  # test a relative path starting from ~
                "~/../../../../../.." + path2 + "/x",
                "foo",
            )  # test a relative path starting from ~
            log.debug("pre-loading: %s", list(pm.get_plugins()))

            assert "testplugin1" not in list(pm.get_plugins())
            assert "testplugin2" not in list(pm.get_plugins())
            pm.reload_plugins()
            log.debug("post-loading: %s", list(pm.get_plugins()))
            assert "testplugin1" in list(pm.get_plugins())
            assert "testplugin2" in list(pm.get_plugins())


def test_load_invalid_plugin(dummy_env, temp_user):
    import evaluation_system.api.plugin_manager as pm

    # this intentionally does not define the required properties and methods
    basic_plugin = textwrap.dedent(
        """
        from sys import modules
        plugin = modules['evaluation_system.api.plugin']
        parameters = modules['evaluation_system.api.parameters']

        class TestPlugin(plugin.PluginAbstract):
            pass
        """
    )

    with tempfile.TemporaryDirectory(prefix="dyn_plugin") as plugin_dir:
        os.makedirs(os.path.join(plugin_dir, "a"))
        with open(plugin_dir + "/a/__init__.py", "w"):
            pass
        with open(plugin_dir + "/a/test_plugin.py", "w") as f:
            f.write(basic_plugin)

        os.environ[pm.PLUGIN_ENV] = "%s,%s" % (
            plugin_dir + "/a",
            "test_plugin",  # test a relative path starting from ~
        )
        log.debug("pre-loading: %s", list(pm.get_plugins()))
        assert "testplugin" not in list(pm.get_plugins())
        pm.reload_plugins()
        log.debug("post-loading: %s", list(pm.get_plugins()))
        assert "testplugin" not in list(pm.get_plugins())


def test_get_plugin_dict(dummy_env):
    import evaluation_system.api.plugin_manager as pm

    with pytest.raises(pm.PluginManagerException):
        pm.get_plugin_metadata("Not available")
    pl = pm.get_plugin_metadata("DummyPlugin")
    assert isinstance(pl, PluginMetadata)
    assert pl.plugin_class == "DummyPlugin"


def test_preview_generation(dummy_env):
    import evaluation_system.api.plugin_manager as pm
    import evaluation_system.misc.config as config

    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.pdf")
        s = os.path.dirname(__file__) + "/test_output/vecap_test_output.pdf"
        pm._preview_copy(s, d)
        assert os.path.isfile(d)
    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.png")
        s = os.path.dirname(__file__) + "/test_output/test_image.png"
        f = open(d, "w")
        f.close()
        pm._preview_copy(s, d)
        assert os.path.isfile(d)
    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.png")
        s = os.path.dirname(__file__) + "/test_output/test_image.png"
        pm._preview_convert(s, d)
        assert os.path.isfile(d)
    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.gif")
        s = os.path.dirname(__file__) + "/test_output/test_animation.gif"
        pm._preview_convert(s, d)
        assert os.path.isfile(d)

    r = pm._preview_generate_name("dummy", {})
    assert "dummy" in r
    assert len(r) == 14
    ts = time.time()
    r = pm._preview_generate_name("murcss", {"timestamp": ts})
    assert "murcss" in r
    assert datetime.datetime.fromtimestamp(ts).strftime("%Y%m%d_%H%M%S") in r

    u = pm._preview_unique_file("murcss", "pdf", {"timestamp": ts})
    assert datetime.datetime.fromtimestamp(ts).strftime("%Y%m%d_%H%M%S") in u

    assert datetime.datetime.fromtimestamp(ts).strftime("%Y%m%d_%H%M%S") in u
    assert "murcss" in u
    assert config.get("preview_path") in u

    r1 = os.path.dirname(__file__) + "/test_output/vecap_test_output.pdf"
    r2 = os.path.dirname(__file__) + "/test_output/test_image.png"
    result = {r1: {"todo": "copy"}, r2: {"todo": "convert"}}
    res = pm._preview_create("murcss", result)
    for r in res:
        assert os.path.isfile(r)
        os.remove(r)


def test_get_command_string(dummy_env, django_user):
    import evaluation_system.api.plugin_manager as pm
    from evaluation_system.model.history.models import History

    h = History.objects.create(
        timestamp=datetime.datetime.now(),
        status=History.processStatus.running,
        uid=django_user,
        configuration='{"some": "config", "dict": "values"}',
        tool="dummytool",
        slurm_output="/path/to/slurm-44742.out",
    )

    cmd = pm.get_command_string(h.id)
    assert "freva-plugin" in cmd
    assert h.tool in cmd

    config = CommandConfig(
        {
            "name": "command",
            "options": "--verbose",
            "tool": "toolname",
            "args": {"list_arg": ["value1", "value2"]},
        }
    )
    expected_string = "command --verbose toolname list_arg=value1,value2"
    assert pm.get_command_string_from_config(config) == expected_string

    config = CommandConfig(
        {
            "name": "command",
            "options": "--verbose",
            "tool": "toolname",
            "args": {"bool_arg": True},
        }
    )
    expected_string = "command --verbose toolname bool_arg=true"
    assert pm.get_command_string_from_config(config) == expected_string


def test_load_scheduled_conf(dummy_env, django_user, temp_user):
    import evaluation_system.api.plugin_manager as pm
    from evaluation_system.model.history.models import History

    h = History.objects.create(
        timestamp=datetime.datetime.now(),
        status=History.processStatus.scheduled,
        uid=django_user,
        configuration='{"some": "config", "dict": "values"}',
        tool="dummytool",
        slurm_output="/path/to/slurm-44742.out",
    )

    res = pm.load_scheduled_conf("dummytool", h.id, temp_user)
    assert res == {}


def test_2dict_to_conf(dummy_env, dummy_plugin):
    import evaluation_system.api.plugin_manager as pm

    configuration = {"number": 1, "the_number": 2, "something": "test"}
    with pytest.raises(pm.PluginManagerException):
        pm.dict2conf("dummytool", configuration)
    with pytest.raises(pm.ParameterNotFoundError):
        pm.dict2conf("dummyplugin", {"some": "config"})
    dd = pm.dict2conf("dummyplugin", configuration)
    assert isinstance(dd, list)
    assert len(dd) == len(configuration)


def test_scheduletool(dummy_env, dummy_plugin):
    import evaluation_system.api.plugin_manager as pm

    with TemporaryDirectory() as td:
        pm.schedule_tool("dummyplugin", log_directory=str(Path(td) / "tmp"))


def test_get_config_name():
    import evaluation_system.api.plugin_manager as pm

    res = pm.get_config_name("dummypluginZ")
    assert res == None


# def test_get_error_warning():
#     import evaluation_system.api.plugin_manager as pm

#     def mock_get_plugin(plugin_name, key, default):
#         return {
#             "error_file": "error.txt",
#             "error_message": "",
#             "warning_file": "warning.txt",
#             "warning_message": "",
#         }.get(key, "")

#     # Files exist
#     with (
#         patch(
#             "evaluation_system.api.plugin_manager.config.get_plugin",
#             side_effect=mock_get_plugin,
#         ),
#         patch(
#             "builtins.open",
#             side_effect=[
#                 MagicMock(read=MagicMock(return_value="Error content")),
#                 MagicMock(read=MagicMock(return_value="Warning content")),
#             ],
#         ),
#     ):
#         error_message, warning_message = pm.get_error_warning("dummyplugin")
#         print(error_message, warning_message)
#         # skip assert for now, because it's already tested in the print statement

#     # Files do not exist
#     with (
#         patch(
#             "evaluation_system.api.plugin_manager.config.get_plugin",
#             side_effect=mock_get_plugin,
#         ),
#         patch("builtins.open", side_effect=Exception("File not found")),
#         patch("evaluation_system.api.plugin_manager.log.warning") as mock_log_warning,
#     ):
#         error_message, warning_message = pm.get_error_warning("dummyplugin")
#         print(error_message, warning_message)
#         # skip assert for now, because it's already tested in the print statement

#     # ConfigurationException
#     with (
#         patch(
#             "evaluation_system.api.plugin_manager.config.get_plugin",
#             side_effect=ConfigurationException("Config error"),
#         ),
#         patch("evaluation_system.api.plugin_manager.log.warning") as mock_log_warning,
#     ):
#         error_message, warning_message = pm.get_error_warning("dummyplugin")
#         print(error_message, warning_message)
#         assert (
#             error_message == ""
#         ), "Error message should be empty on ConfigurationException."
#         assert (
#             warning_message == ""
#         ), "Warning message should be empty on ConfigurationException."


# def test_get_plugin_version():
#     import evaluation_system.api.plugin_manager as pm

#     mock_get_plugins = {
#         "existing_plugin": MagicMock(plugin_module="path/to/existing_plugin_module"),
#     }
#     mock_repository_get_version = MagicMock(
#         return_value=("https://freva_repository.url", "git_hash_value")
#     )

#     # Plugin is the module itself
#     with (
#         patch("evaluation_system.api.plugin_manager.__version_cache", {}),
#         patch(
#             "evaluation_system.api.plugin_manager.get_plugins",
#             return_value=mock_get_plugins,
#         ),
#         patch("inspect.getfile", return_value="path/to/self_module"),
#         patch(
#             "evaluation_system.model.repository.get_version",
#             mock_repository_get_version,
#         ),
#     ):

#         repository_url, git_hash = pm.get_plugin_version("self")
#         assert (
#             repository_url == "https://freva_repository.url"
#         ), "Repository URL mismatch for 'self'."
#         assert git_hash == "git_hash_value", "Git hash mismatch for 'self'."

#     # Plugin not found
#     with (
#         patch("evaluation_system.api.plugin_manager.__version_cache", {}),
#         patch(
#             "evaluation_system.api.plugin_manager.get_plugins",
#             return_value=mock_get_plugins,
#         ),
#     ):
#         try:
#             pm.get_plugin_version("non_existing_plugin")
#             assert False, "Expected PluginManagerException for non-existing plugin."
#         except pm.PluginManagerException as e:
#             assert (
#                 str(e) == "Plugin <non_existing_plugin> not found"
#             ), "Incorrect exception message."



import logging
import os
import sys
from typing import Optional, Union, cast

from rich.console import Console
from rich.logging import RichHandler

logger_stream_handle = RichHandler(
    rich_tracebacks=True,
    show_path=False,
    console=Console(soft_wrap=False, stderr=True),
)
logger_stream_handle.setLevel(logging.INFO)
logging.basicConfig(
    format="%(name)s - %(levelname)s - %(message)s",
    handlers=[logger_stream_handle],
    datefmt="[%X]",
    level=logging.INFO,
)


class FrevaLogger(logging.Logger):
    """Custom logger to assure that all log handles receive the same level."""

    is_cli: bool = False
    """Indicate whether or not this logger belongs to a cli process."""

    def setLevel(self, level: Union[int, str]) -> None:
        super().setLevel(level)
        for handler in self.handlers:
            handler.setLevel(level)

    def set_level(self, level: Union[int, str]) -> None:
        """Set the log level of the logger."""
        self.setLevel(level)


logging.setLoggerClass(FrevaLogger)
logger: FrevaLogger = cast(FrevaLogger, logging.getLogger("freva"))
logger.setLevel(logging.INFO)
logger.propagate = False
logger.addHandler(logger_stream_handle)


class _ConfigWrapper:
    """Convenience class that helps to dynamically set the location of the
    evaluation system config file."""

    _env: str = "EVALUATION_SYSTEM_CONFIG_FILE"

    def __init__(
        self,
        default_file: str,
    ):
        self.default_file = default_file

    def __fspath__(self) -> str:
        return os.environ.get(self._env, self.default_file)

    def __repr__(self) -> str:
        return self.__fspath__()



import logging
import os
import warnings

from evaluation_system import __version__
from evaluation_system.misc import logger

from ._databrowser import count_values, databrowser, facet_search
from ._esgf import esgf_browser, esgf_datasets, esgf_download, esgf_facets, esgf_query
from ._history import history
from ._plugin import (
    get_tools_list,
    list_plugins,
    plugin_doc,
    plugin_info,
    read_plugin_cache,
    run_plugin,
)
from ._user_data import UserData
from .utils import PluginStatus, config, is_jupyter

warnings.filterwarnings("always", category=DeprecationWarning, module="freva.*")

if is_jupyter():
    os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
    logger.setLevel(logging.WARNING)

__all__ = [
    "__version__",
    "config",
    "UserData",
    "databrowser",
    "count_values",
    "facet_search",
    "run_plugin",
    "list_plugins",
    "plugin_info",
    "plugin_doc",
    "esgf_browser",
    "esgf_facets",
    "esgf_datasets",
    "esgf_download",
    "esgf_query",
    "history",
    "PluginStatus",
]



"""This module queries the database for plugin history entries."""

from __future__ import annotations

import json
from typing import Any, Optional, Union

import lazy_import
from django.core.exceptions import ImproperlyConfigured

from evaluation_system.misc import logger
from evaluation_system.misc.config import reloadConfiguration

db_settings = lazy_import.lazy_module("evaluation_system.settings.database")
pm = lazy_import.lazy_module("evaluation_system.api.plugin_manager")
from .utils import config

try:
    from evaluation_system.model.user import User
except ImproperlyConfigured:
    reloadConfiguration()
    db_settings.reconfigure_django()


__all__ = ["history"]


def history(
    *args: str,
    limit: int = 10,
    plugin: Optional[str] = None,
    since: Optional[str] = None,
    until: Optional[str] = None,
    entry_ids: Union[int, list[int], None] = None,
    full_text: bool = False,
    return_results: bool = False,
    return_command: bool = False,
    _return_dict: bool = True,
    user_name: Optional[str] = None,
) -> Union[list[Any], dict[str, Any]]:
    """Get access to the configurations of previously applied freva plugins.

    The `.history` method displays the entries with a one-line compact description.
    The first number you see is the entry id, which you might use to select
    single entries.

    Parameters
    ----------
    limit: int, default: 10
      Limit the number of entries to be displayed.
    plugin: str, default: None
      Display only entries from a given plugin name.
    since: str, datetime.datetime, default: None
      Retrieve entries older than date, see hint on date format below.
    until: str, datetime.datetime, default: None
      Retrieve entries younger than date, see hint on date format below.
    entry_ids: list, default: None
       Select entries whose ids are in "ids".
    full_text: bool, default: False
      Show the complete configuration.
    return_results: bool, default: False
      Also return the plugin results.
    return_command: bool, default: False
      Return the commands instead of history objects.
    user_name: str, default: None
      Select entries belonging to another user (e.g., user_name="<username>")
      or all users (all or *).
    _return_dict: bool, default: True
      Return a dictionary representation, this is only for internal use.

    Returns
    -------
      list:
        freva plugin history

    Example
    -------

    Get the last three history entries:

    .. execute_code::

        import freva
        hist = freva.history(limit=3)
        print(type(hist), len(hist))
        print(hist[-1].keys())
        config = hist[-1]['configuration']
        print(config)


    .. hint:: Date Format
        Dates are given in the ISO-8601 format and can be "YYYY-MM-DDTHH:mm:ss.n"
        or any less accurate subset. These are all valid: "2012-02-01T10:08:32.1233431",
        "2012-02-01T10:08:32", "2012-02-01T10:08", "2012-02-01T10", "2012-02-01",
        "2012-02", "2012". Missing values are assumed to be the minimal allowed value.
        For example: "2012" = "2012-01-01T00:00:00.0"
    """
    if not isinstance(entry_ids, (list, tuple, set)) and entry_ids is not None:
        entry_ids = [entry_ids]
    if not return_command:
        prefix = "history"
        if plugin:
            prefix = "history of {plugin}"
        logger.info(
            f"{prefix}, limit={limit}, since={since},"
            f" until={until}, entry_ids={entry_ids}"
        )

    rows = pm.get_history(
        user=None,
        plugin_name=plugin,
        limit=limit,
        since=since,
        until=until,
        entry_ids=entry_ids,
        user_name=user_name,
    )

    if rows:
        # pass some option for generating the command string
        if return_command:
            commands = []
            for row in rows:
                cmd = pm.get_command_string_from_row(row)
                commands.append(cmd)
            return commands
        if _return_dict:
            command_dicts = []
            for cmd_dict in [row.__dict__ for row in rows]:
                out_dict = {}
                for key, value in cmd_dict.items():
                    try:
                        out_dict[key] = json.loads(value)
                    except (json.JSONDecodeError, TypeError):
                        if key in ("_state",):
                            continue
                        if key == "timestamp":
                            out_dict[key] = value.isoformat()
                        else:
                            out_dict[key] = value
                if return_results:
                    out_dict["result"] = pm.get_result_output(cmd_dict["id"])
                else:
                    out_dict["result"] = {}
                command_dicts.append(out_dict)
            return command_dicts
        return rows
    return []
